#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include <alloca.h>
#include "Decomp3D.h"
#include "Greedy.h"
#include "Face.h"
#include "BinFileHandler.h"

#define NEW_GREEDY

//---------------------------------------------------
void Decomp3D::getDecomp(const char* name, int numElem)
{

  int *connect = new int[numElem];
  char buffer[80];
  int i;
  int nele;

  if (name[0] != 0)  {
    ifstream decompFilePtr(name,ios::in);
    // check that file ptr exists
    if(!decompFilePtr){
      fprintf(stderr, "*** Error opening file: %s\n", name);
      fprintf(stderr, "... Aborting\n");
      decompFilePtr.close();
      exit(1);
    }

    // Scan input file for the word Decomposition
    // Then, read in decomp info 
    decompFilePtr.get(buffer[0]);
    i = 0;
    while(buffer[i] != '\n')  {
      i++;
      decompFilePtr.get(buffer[i]);
    }

    while (strncmp(buffer,"Decomposition",13))  {
      decompFilePtr.get(buffer[0]);
      i=0;
      while(buffer[i] != '\n'){
        i++;
        decompFilePtr.get(buffer[i]);
      }
    }

    // check number of subdomains
    decompFilePtr >> numSub;
    if(99999 < numSub || numSub < 0)  {
      fprintf(stderr, "***ERROR: Number of subdomains incorrectly specified: %d\n", numSub);
      fprintf(stderr, "...Aborting\n");
      exit(1);
    }

    // form sub to element connectivity
    int *cx = new int[numSub+1];

    int curEle = 0;
    for(int isub = 0; isub < numSub; ++isub) {
      //fprintf(stderr,"... Reading SubDomain #: %d\n", isub);
      decompFilePtr >> nele;
      cx[isub] = curEle;

      if(curEle + nele > numElem) {
        fprintf(stderr, "***ERROR: Decomposition contains more elements than the original mesh (%d)\n", numElem);
        fprintf(stderr,"...Aborting\n");
        exit(1);
      }

      for(int iele=0; iele < nele; ++iele) {
        decompFilePtr >> connect[curEle];
        connect[curEle++]--;
      }
    }
    cx[numSub] = curEle;
    subToElem = new Connectivity(numSub,cx,connect);
  }

  else  {
    numSub=1;  // default if no decomp file is specified
    int* cx = new int[2];

    cx[0]=0;
    cx[1]=numElem;
    for(i = 0; i < numElem; i++)
      connect[i]=i;
    subToElem = new Connectivity(numSub,cx,connect);
  }

}

//--------------------------------------------------------------------

void Decomp3D::makeConnectivity(int numClus, int numElem, int numNodes, 
				Element **el, Connectivity *elemToClus)
{

  elemToNode = new Connectivity(numElem, el);

  subToNode = subToElem->transcon(elemToNode);

  nodeToSub = subToNode->reverse();

  subToSub = subToNode->transcon(nodeToSub);

  if (elemToClus) {
    Connectivity *subToClus = subToElem->transcon(elemToClus);
    clusToSub = subToClus->reverse();
  }
  else {
    GreedyDecomp *clusDec = new GreedyDecomp("clusterDecomp", numClus);
#ifdef NEW_GREEDY
    long long *cost = new long long[numSub];
    for (int iSub=0; iSub<numSub; ++iSub)
      cost[iSub] = subToElem->num(iSub);
    clusToSub = clusDec->cpuGreedyBuild(subToSub, nodeToSub, subToNode, cost, numSub);
    if (cost) delete [] cost;
#else
    clusToSub = cpuDec->greedyBuild(subToSub, nodeToSub, subToNode);
#endif
  }

  clusToNode = clusToSub->transcon(subToNode);

}

//---------------------------------------------------------------

int Decomp3D::remapNodes(std::pair<int,int> *nodeMap, int iClus)
{

  int numNodes = clusToNode->num(iClus);
/*
  // sort the node to sub connectivity
  int iNode;
  for(iNode = 0; iNode < numNodes; ++iNode)
    std::sort((*nodeToSub)[iNode], (*nodeToSub)[iNode]+nodeToSub->num(iNode));
*/

  // create a mapping between subdomain list to unique id
  std::map<NodeConnect,int> connectSet;

  int uniqueId = 0;  // the unique id
  int actNumNodes = 0;
  int maxSubs = 0;

  int iNode;
  for(iNode = 0; iNode < numNodes; ++iNode)  { 
 
    int clusNode = (*clusToNode)[iClus][iNode];
    // check existence of node
    if(nodeToSub->num(clusNode) == 0) continue;

    NodeConnect thisConnect(nodeToSub->num(clusNode), (*nodeToSub)[clusNode]);

    std::map<NodeConnect,int>::iterator it = connectSet.find(thisConnect);

    if ( it == connectSet.end() )   // this set was not yet found
      connectSet[thisConnect] = uniqueId++;

    nodeMap[actNumNodes++] = std::pair<int,int>(connectSet[thisConnect], clusNode);
  }

   std::stable_sort(nodeMap, nodeMap+actNumNodes);
   
   // loop over nodes in nodeMap to count number of subdomains
   for (iNode = 0; iNode < actNumNodes; ++iNode)
     if ( nodeToSub->num( nodeMap[iNode].second ) > maxSubs )
       maxSubs++;

   return actNumNodes;
}

//----------------------------------------------------------------------

void Decomp3D::getDomainConnect() {
 
  // get number of clusters
  int numClus = clusToNode->csize();

  // create information on subdomain neighbors
  // and allocate memory for numInterfNodes array

  int numSubs = subToSub->csize();
  numSubNeighbors = new int[numSubs];
  connectedDomain = new int *[numSubs];
  // loop through subtosub connectivity and record
  // the neighbors to each subdomain
  int iSub;
  for (iSub = 0; iSub < numSubs; iSub++)  {

    int numLocSubs = subToSub->num(iSub);
    numSubNeighbors[iSub] = numLocSubs - 1;

    int iConn = 0;

    // allocate memory for each connected Domain array
    connectedDomain[iSub] = new int[numSubNeighbors[iSub]];

    // populate those subs into connectedDomain array
    for (int jSub = 0; jSub < numLocSubs; jSub++)  {
      
      int thisSub = (*subToSub)[iSub][jSub];
      if ( thisSub != iSub )
        connectedDomain[iSub][iConn++] = thisSub;
    }
    std::stable_sort(connectedDomain[iSub], connectedDomain[iSub]+numSubNeighbors[iSub]);
  }

  // allocate memory for interface nodes according to # of subs
  interfNodes = new Connectivity *[numSubs];

  // count # of interface nodes between each subdomain
  for (iSub = 0; iSub < numSubs; iSub++)  {
    // get number of nodes in iSub
    int numLocSubNodes = subToNode->num(iSub);

    int numInterfs = 0; 

    // get number of interfaces including their multiplicity
    int iNode;
    for (iNode = 0; iNode < numLocSubNodes; iNode++)
      numInterfs += nodeToSub->num((*subToNode)[iSub][iNode]) - 1;

    // create pair of unique id and node number
    std::pair<int,int> *sub2InterfNode = new std::pair<int,int>[numInterfs];

    // populate sub2InterfNode
    int count = 0;
    for (iNode = 0; iNode < numLocSubNodes; iNode++)  {
      
      // get global node number
      int thisNode = (*subToNode)[iSub][iNode];

      for (int iConn = 0; iConn < nodeToSub->num(thisNode); iConn++)  {
	if ((*nodeToSub)[thisNode][iConn] != iSub)  {
	  sub2InterfNode[count++] = std::pair<int,int>
			((*nodeToSub)[thisNode][iConn], thisNode);
   	}
      }
    }

    // sort sub2InterfNode by subdomain #, then node number
    std::stable_sort(sub2InterfNode, sub2InterfNode+numInterfs);

    // create arrays necessary for creating interface node connectivity
    int *ptrs = new int[ numSubNeighbors[iSub]+1 ];
    int *nn = new int[numInterfs];

    int neighborCount = 0;
    int curSub = -1;

    for (iNode = 0; iNode < numInterfs; iNode++)  {

      nn[iNode] = sub2InterfNode[iNode].second;

      // check if next subdomain interface is reached
      if (sub2InterfNode[iNode].first > curSub)  {
        ptrs[neighborCount++] = iNode;
        curSub = sub2InterfNode[iNode].first;
      }
    }

    ptrs[neighborCount] = numInterfs;  

    if (neighborCount != numSubNeighbors[iSub])
      fprintf(stderr,"Bad neighbor count: %d vs %d\n", 
			neighborCount, numSubNeighbors[iSub]);


    interfNodes[iSub] = new Connectivity(numSubNeighbors[iSub], ptrs, nn);
  }
}

//--------------------------------------------------------------------

void Decomp3D::writeCpuToSubMap(int numCpus, const char *outPrefix, int type)  
{

  GreedyDecomp* cpuDec = new GreedyDecomp("cpusDecomp", numCpus);

  Connectivity* cpuToSub;

  if (clusToSub->csize() == 1) {
#ifdef NEW_GREEDY
    long long *cost = new long long[numSub];
    for (int iSub=0; iSub<numSub; ++iSub)
      cost[iSub] = subToElem->num(iSub);
    cpuToSub = cpuDec->cpuGreedyBuild(subToSub, nodeToSub, subToNode, cost, numSub);
    if (cost) delete [] cost;
#else
    cpuToSub = cpuDec->greedyBuild(subToSub, nodeToSub, subToNode);
#endif
  }
  else if(clusToSub->csize() == numCpus) {
    cpuToSub = clusToSub;
  }
  else {
    int* pointer = new int[numCpus+1];
    int* target = new int[numSub];
    cpuToSub = new Connectivity(numCpus, pointer, target);
    //int nSubPerCpu = numSub/numCpus;  // NOT ACCEPTABLE!!!
    int *nSubPerCpu = new int[numCpus]; // PJSA
    int i;
    for(i=0; i<numCpus; ++i) 
      nSubPerCpu[i] = (i < numSub%numCpus) ? numSub/numCpus+1 : numSub/numCpus;
    int count = 0;
    for (i=0; i<clusToSub->csize(); ++i) {
      for (int j=0; j<clusToSub->num(i); ++j) {
	target[count++] = (*clusToSub)[i][j];
      }
    }
    // for (i=0; i<numCpus+1; ++i) pointer[i] = i*nSubPerCpu;
    pointer[0] = 0;
    for (i=0; i<numCpus; ++i) pointer[i+1] = pointer[i] + nSubPerCpu[i]; // PJSA
  }

  char cpuFileName[MAXLINE];
  sprintf(cpuFileName, "%s.%dcpu", outPrefix, numCpus);

  fprintf(stdout, "... Writing \'%s\'\n", cpuFileName);

  FILE *cpuFile = fopen(cpuFileName, "w");

  fprintf(cpuFile, "%d\n", numCpus);

  for (int iCpu=0; iCpu<numCpus; ++iCpu) {

    int numLocalSub = cpuToSub->num(iCpu);

    if (type > 3)
      fprintf(cpuFile, "  %d\n", numLocalSub);
    else
      fprintf(cpuFile, "  %d %d\n", numLocalSub, 1);

    for (int iSub=0; iSub<numLocalSub; ++iSub)
      fprintf(cpuFile, "%d\n", (*cpuToSub)[iCpu][iSub] + 1);

  }

  fclose(cpuFile);

}

//----------------------------------------------------------------

int *Decomp3D::printElements(int iClus, BinFileHandler &outFileGeo, int &numClusElems, 
			     Element **elements, MapVec &gl2ClElems)  
{

  int numSubs = clusToSub->num(iClus);

  // count number of elements in cluster

  numClusElems = 0;

  int iSub;
  for (iSub = 0; iSub < numSubs; iSub++)
    numClusElems += subToElem->num( (*clusToSub)[iClus][iSub] );

  int *elemPtrs = new int[numClusElems];

  int count = 0;
  int numInts = 0;
  for (iSub = 0; iSub < numSubs; iSub++) {
 
    int subD = (*clusToSub)[iClus][iSub];
    // get # of elems for current subdomain
    int numEle = subToElem->num(subD);

    for (int iElem = 0; iElem < numEle; iElem++) {

      // get next element in current subdomain
      int elemI = (*subToElem)[subD][iElem];
      gl2ClElems[elemI] = count;

      // get number of nodes for each element
      elemPtrs[count] = numInts;
      count++;
      // increment numInts by # elem nodes, elemtype, gl. elem #, volume id
      numInts += elements[elemI]->numNodes() + 3;
    }
  }

  int *elementData = new int[numInts];

  count = 0;
  for (iSub = 0; iSub < numSubs; iSub++)  {

    int subD = (*clusToSub)[iClus][iSub];
    // get # of elems for current subdomain
    int numEle = subToElem->num(subD);
    //fprintf(stderr,"Sub %d has %d elems\n", subD, numEle);

    // populate elementdata array
    for (int iElem = 0; iElem < numEle; iElem++)  {
      // get next element in current subdomain
      int elemI = (*subToElem)[subD][iElem];

      // put type in array
      elementData[ elemPtrs[count] ] = elements[elemI]->getType();

      // put global elem number in array
      elementData[ elemPtrs[count] + 1 ] = elemI;

      // put volume id in array (porous media)
      elementData[ elemPtrs[count] + 2 ] = elements[elemI]->getVolumeID(); 

      // put local node numbers of element in array
      elements[elemI]->nodes(elementData+elemPtrs[count] + 3);
      count++;
    }
  }
 
  // bin write out big array
  outFileGeo.write(elementData, numInts);
 
  delete [] elementData;

  return elemPtrs;

}

//----------------------------------------------------------------

void Decomp3D::printElements(int iClus, 
			     EnsightFileHandler &outFileGeo, 
			     EnsightFileHandler &outFileMat, 
			     int &numClusElems, Element **elements, 
			     MapVec &gl2ClElems)  
{

  int numSubs = clusToSub->num(iClus);

  // count number of elements in cluster
  numClusElems = 0;

  int iSub;
  for (iSub = 0; iSub < numSubs; iSub++)
    numClusElems += subToElem->num( (*clusToSub)[iClus][iSub] );

  int nTet = 0, nPrism = 0;
  int count = 0;

  for (iSub = 0; iSub < numSubs; iSub++) {
 
    int subD = (*clusToSub)[iClus][iSub];

    // get # of elems for current subdomain
    int numEle = subToElem->num(subD);
    for (int iElem = 0; iElem < numEle; iElem++) {

      // get next element in current subdomain
      int elemI = (*subToElem)[subD][iElem];
      gl2ClElems[elemI] = count;
      count++;

      if(elements[elemI]->getType() == 5)
        nTet++;
      else
        nPrism++;      
    }
  }

  int *tetId      = new int[nTet];
  int *tetNodes   = new int[nTet*4];;
  int *prismId    = new int[nPrism];
  int *prismNodes = new int[nPrism*6];
  
  nTet  = nPrism = count = 0;
  for (iSub = 0; iSub < numSubs; iSub++)  {

    int subD = (*clusToSub)[iClus][iSub];

    // get # of elems for current subdomain
    int numEle = subToElem->num(subD);

    for (int iElem = 0; iElem < numEle; iElem++)  {
      // get next element in current subdomain
      int elemI = (*subToElem)[subD][iElem];

      if(elements[elemI]->getType() == 5) {
	// put global elem number in array
	tetId[nTet] = elemI;
	
	// put local node numbers of element in array
	elements[elemI]->nodes(tetNodes + 4*nTet);

	// Ensight numbering starts at 1
	for (int k=0; k<4; k++)
	  tetNodes[4*nTet+k]++;
	
        nTet++;
      } else {
	// put global elem number in array
	prismId[nPrism] = elemI;
	
	// put local node numbers of element in array
	elements[elemI]->nodes(prismNodes + 6*nPrism);
	
	// Ensight numbering starts at 1
	for (int k=0; k<6; k++)
	  prismNodes[6*nPrism+k]++;

        nPrism++;
      }

      count++;
    }
  }
 
  fprintf(stderr, "Num tets: %d Num prisms: %d\n", 
	  nTet, nPrism);

  // bin write out tetras
  if (nTet>0) {
    outFileGeo.write("tetra4");
    outFileGeo.write(&nTet,    1);
    outFileGeo.write(tetId,    nTet, 1);
    outFileGeo.write(tetNodes, nTet, 4);

    for (int i=0; i<nTet; i++)
      tetId[i] = 4;
    outFileMat.write("tetra4");
    outFileMat.write(tetId, nTet, 1);
  }

  // bin write out prisms
  if (nPrism>0) {
    outFileGeo.write("penta6");
    outFileGeo.write(&nPrism,    1);
    outFileGeo.write(prismId,    nPrism, 1);
    outFileGeo.write(prismNodes, nPrism, 6);

    for (int i=0; i<nPrism; i++)
      prismId[i] = 5;
    outFileMat.write("penta6");
    outFileMat.write(prismId, nPrism, 1);
  }
 
  delete [] tetId;
  delete [] tetNodes;
  delete [] prismId;
  delete [] prismNodes;

}

//--------------------------------------------------------------

// WARNING: This function works ONLY for fluid Tetrahedrons

void Decomp3D::makeSubToFace(int numFaces, int numElem, 
					Face *faces)  {

  int eleCount[3];
  int *eleList[3];
 
  int* connect = new int[numElem];
  int* cx = new int[numFaces + 1];

  Connectivity *nodeToElem = elemToNode->reverse();

  int count = 0;
  for (int i = 0; i < numFaces; i++) {

    for (int fNode = 0; fNode < 3; fNode++)  {

      cx[i] = count;

      // get a node from face
      int thisNode = faces[i][fNode];

      // count number of elements connected to thisNode
      eleCount[fNode] = nodeToElem->num(thisNode);

      //assign pointer list of elems connected to thisNode
      eleList[fNode] = (*nodeToElem)[thisNode];
    }

    // loop over the attached elements to each node
    // and test for a matching element with the other face nodes
    for (int j = 0; j < eleCount[0]; j++)
      for (int k = 0; k < eleCount[1]; k++)
        if (eleList[0][j] == eleList[1][k]) {

          for (int l = 0; l < eleCount[2]; l++)
            if (eleList[1][k] == eleList[2][l]) {
              connect[count++] = eleList[0][j];
              goto end; // escape from loop when match is made
            }

        }
    end: ;

  }

  cx[numFaces] = count;

  Connectivity *faceToElem = new Connectivity(numFaces, cx, connect);
  Connectivity *elemToFace = faceToElem->reverse();
  subToFace = subToElem->transcon(elemToFace);

}

//--------------------------------------------------------------

void Decomp3D::makeSubToMatch(Connectivity *subToData, int numMatch, int *matchData)
{

  int i, iSub, numSubs = subToData->csize();

  MapVec gl2GlMatchMap;

  for (i=0; i<numMatch; ++i)
    gl2GlMatchMap[ matchData[i] ] = i;

  int *ptrs = new int[numSubs + 1];
 
  int numTargets = 0;
  for (iSub=0; iSub<numSubs; ++iSub) {
    
    ptrs[iSub] = numTargets;

    for (i=0; i<subToData->num(iSub); ++i)
      if (gl2GlMatchMap.find((*subToData)[iSub][i]) != gl2GlMatchMap.end())
	numTargets++;

  }

  ptrs[numSubs] = numTargets;
  int *targets = new int[numTargets];

  int count = 0;
  for (iSub=0; iSub<numSubs; ++iSub) {

    for (i=0; i<subToData->num(iSub); ++i)
      if (gl2GlMatchMap.find((*subToData)[iSub][i]) != gl2GlMatchMap.end())
        targets[count++] = gl2GlMatchMap[ (*subToData)[iSub][i] ];

  }

  for (iSub=0; iSub<numSubs; ++iSub)
    stable_sort(targets + ptrs[iSub], targets + ptrs[iSub + 1]);

  subToMatch = new Connectivity(numSubs, ptrs, targets);

  clusToMatch = clusToSub->transcon(subToMatch);

  //clusToMatchPts->print();

  //subToMatchPts->print();

}

//--------------------------------------------------------------

void Decomp3D::makeSubToMatch(int numData, Connectivity *subToData, int numMatch, int *matchData)
{

  int i, iSub, numSubs = subToData->csize();

  int *mycount = new int[numData];
  int *myptrs = new int[numData + 1];
  int *mytargets = new int[numMatch];
  
  for (i=0; i<numData; ++i)
    mycount[i] = 0;

  for (i=0; i<numMatch; ++i)
    mycount[ matchData[i] ] += 1;

  myptrs[0] = 0;
  for (i=1; i<numData+1; ++i)
    myptrs[i] = myptrs[i-1] + mycount[i-1];

  for (i=0; i<numData; ++i)
    mycount[i] = myptrs[i];

  for (i=0; i<numMatch; ++i) {
    int gl = matchData[i];
    mytargets[ mycount[gl] ] = i;
    mycount[gl] += 1;
  }

  if (mycount) delete [] mycount;

  Connectivity glToMatch(numData, myptrs, mytargets);
  
  //glToMatch.print();

  int *ptrs = new int[numSubs + 1];
 
  int numTargets = 0;
  for (iSub=0; iSub<numSubs; ++iSub) {
    
    ptrs[iSub] = numTargets;

    for (i=0; i<subToData->num(iSub); ++i) {
      int gl = (*subToData)[iSub][i];
      numTargets += glToMatch.num(gl);
    }

  }

  ptrs[numSubs] = numTargets;
  int *targets = new int[numTargets];

  int count = 0;
  for (iSub=0; iSub<numSubs; ++iSub) {
    for (i=0; i<subToData->num(iSub); ++i) {
      int gl = (*subToData)[iSub][i];
      for (int k=0; k<glToMatch.num(gl); ++k)
	targets[count++] = glToMatch[gl][k];
    }
  }

  for (iSub=0; iSub<numSubs; ++iSub)
    stable_sort(targets + ptrs[iSub], targets + ptrs[iSub + 1]);

  subToMatch = new Connectivity(numSubs, ptrs, targets);

  clusToMatch = clusToSub->transcon(subToMatch);

  //clusToMatch->print();

  //subToMatch->print();

}

