#include "Domain.h"

#include "Tetrahedron3D.h"
#include "BinFileHandler.h"
#include "EnsightFileHandler.h"

#include <cstdio>
#include <cstdlib>

#ifdef OLD_STL
#include <stack.h>
#include <map.h>
#else
#include <stack>
#include <map>
using std::stack;
using std::map;
#endif

//--------------------------------------------------------------------

FluidDomain::FluidDomain(int numClus) : Domain(numClus)  
{

  numFace = 0;
  numMatchPts = 0;

  type_active = new bool[MAX_TYPE_BC - MIN_TYPE_BC + 1];
  type_name = new const char*[MAX_TYPE_BC - MIN_TYPE_BC + 1];

  type_active -= MIN_TYPE_BC;
  type_name -= MIN_TYPE_BC;

  int i;
  for (i=MIN_TYPE_BC; i<=MAX_TYPE_BC; ++i) {
    type_active[i] = false;
    type_name[i] = "InactiveSurface";
  }

  type_active[-5] = true;
  type_active[-4] = true;
  type_active[-3] = true;
  type_active[-2] = true;
  type_active[2] = true;
  type_active[3] = true;
  type_active[4] = true;
  type_active[5] = true;
  type_active[6] = true;

  type_name[-5] = "OutletMoving";
  type_name[-4] = "InletMoving";
  type_name[-3] = "StickMoving";
  type_name[-2] = "SlipMoving";
  type_name[0] = "Internal";
  type_name[2] = "SlipFixed";
  type_name[3] = "StickFixed";
  type_name[4] = "InletFixed";
  type_name[5] = "OutletFixed";
  type_name[6] = "Symmetry"; // PJSA
  type_name[10] = "Missing";

}

//--------------------------------------------------------------------

void FluidDomain::getGeometry(const char* file)  
{

  FILE* fp = fopen(file, "r");
  if (!fp) {
    fprintf(stderr, "Error: could not open file \'%s\'\n", file);
    exit(1);
  }
  char line[MAXLINE], keyword[MAXLINE];
  fgets(line, MAXLINE, fp);
  sscanf(line, "%s", keyword);
  int type = 0;
  if (strcmp(keyword, "Nodes") == 0 || 
      strcmp(keyword, "Elements") == 0 ||
      strcmp(keyword, "typedef") == 0)
    type = 1;
  else if (strcmp(keyword, "FIELDVIEW") == 0)
    type = 2;
  else
    type = 3;
  fclose(fp);

  if (type == 1)
    getGeometryXpost(file);
  else if (type == 2)
    getGeometryFieldView(file);
  else if (type == 3)
    getGeometrySinus(file);

}

//--------------------------------------------------------------------

struct SNode {
  double x[3];
};
struct SBFace {
  int nodes[3];
  int type;
  int surface_id;
};
struct SElement {
  int nodes[4];
  int volume_id;
  public:
  SElement() { volume_id = -1;}
  ~SElement() {}
};
struct SBCond {
  int node;
  int dof;
  double val;
};

//------------------------------------------------------------------------------

void FluidDomain::getGeometryXpost(const char *meshname)  
{

  map<int,int> type_map;
  type_map[4] = 4;
  type_map[5] = 5;

  stack<SNode> snodes;
  stack<SBFace> sbfaces;
  stack<SElement> selements;
  stack<SBCond> sdbcs;

  FILE* fp = fopen(meshname, "r");
  if (!fp) {
    fprintf(stderr, "Error: could not open file \'%s\'\n", meshname);
    exit(1);
  }

  int type_boco = 10;
  int type_read = 0;
  int surface_id = 0;
  int volume_id = -1;
  char line[MAXLINE], key1[MAXLINE], key2[MAXLINE], dummy[MAXLINE];
  while (fgets(line, MAXLINE, fp) != 0) {
    sscanf(line, "%s", key1);
    bool skip = false;
    if (strcmp(key1, "Nodes") == 0) {
      skip = true;
      type_boco = 10;
      type_read = 1;
      sscanf(line, "%*s %s", key2);
    }
    else if (strcmp(key1, "Elements") == 0) {
      skip = true;
      type_boco = 10;
      type_read = 2;
      sscanf(line, "%*s %s", key2);

      // Fluid Mesh
      if (strncmp(key2,"FluidMesh", strlen("FluidMesh")) == 0) {
        int k = strlen("FluidMesh");
        int underscore_pos = -1;
        while((key2[k] != '\0') && (k<MAXLINE)) {
          if(key2[k] == '_') underscore_pos = k;
          k++;
        }
        if(underscore_pos > -1)
          sscanf(key2+(underscore_pos+1),"%d", &volume_id);
      }

      // CBM: porous
      if (strncmp(key2,"PorousVolume", strlen("PorousVolume")) == 0) {
        int k=strlen("PorousVolume");
        int underscore_pos = -1;
        while((key2[k] != '\0') && (k<MAXLINE)) {
          if(key2[k] == '_') underscore_pos = k;
          k++;
        }
        if(underscore_pos > -1) 
          sscanf(key2+(underscore_pos+1),"%d",&volume_id);
      }


      for (int j=MIN_TYPE_BC; j<=MAX_TYPE_BC; ++j) {
	if (strncmp(key2, type_name[j], strlen(type_name[j])) == 0) {
	  type_boco = j;
          surface_id = 0; //initialization

          // new idea: read the surface id after the last underscore in the element set name
          int k=strlen(type_name[j]);
          int underscore_pos = -1;
          while((key2[k] != '\0') && (k<MAXLINE)) {
            if(key2[k] == '_') underscore_pos = k;
            k++;
          }
          if(underscore_pos > -1)  {
            sscanf(key2+(underscore_pos+1),"%d",&surface_id); 

            if (surface_id > 0)  {
              char pt1[MAXLINE];
              char pt2[MAXLINE];
              char pt3[MAXLINE];
              char pt4[MAXLINE];
              char pt5[MAXLINE];
              char pt6[MAXLINE];
              sscanf(line, "%*s %*s %*s %*s %s %s %s %s %s %s", pt1, pt2, pt3, pt4, pt5, pt6);
            }
          }

          //sscanf(key2+strlen(type_name[j])+8,"%d",&surface_id); // PJSA
	  if (type_boco == -1 || type_boco == 1) 
	    fprintf(stderr, "WARNING: POSSIBLE BAD boco: %d\n", type_boco);
	  break;
	}
      }
     
    }
    else if (strcmp(key1, "typedef") == 0) {
      skip = true;
      int t;
      sscanf(line, "%*s %d %s", &t, key2);
      if (strcmp(key2, "Tri_3") == 0)
	type_map[t] = 4;
      else if (strcmp(key2, "Tetra_4") == 0)
	type_map[t] = 5;
    }
    else if (strcmp(key1, "SDBoundary") == 0) { // PJSA 6-28-05
      skip = true;
      type_read = 3;
      type_boco = 10;
      sscanf(line, "%*s %s", key2);
    }

    if (!skip) {
      if (type_read == 1) {
	SNode sn;
	sscanf(line, "%*s %lf %lf %lf", &sn.x[0], &sn.x[1], &sn.x[2]);
	snodes.push(sn);
      }
      else if (type_read == 2) {
	int num, t;
	sscanf(line, "%d %d", &num, &t);
	map<int,int>::iterator it = type_map.find(t);
	if (it != type_map.end())
	  t = it->second;

	if (t == 4) {
	  if (type_active[type_boco]) {
	    SBFace sbf;
	    sscanf(line, "%*s %*s %d %d %d", &sbf.nodes[0], &sbf.nodes[1], &sbf.nodes[2]);
	    sbf.nodes[0] -= 1;
	    sbf.nodes[1] -= 1;
	    sbf.nodes[2] -= 1;
	    sbf.type = type_boco;
            sbf.surface_id = surface_id; // PJSA
	    sbfaces.push(sbf);
	  }
	}
	else if (t == 5) {
	  SElement se;
	  sscanf(line, "%*s %*s %d %d %d %d", &se.nodes[0], &se.nodes[1], &se.nodes[2], &se.nodes[3]);
	  se.nodes[0] -= 1;
	  se.nodes[1] -= 1;
	  se.nodes[2] -= 1;
	  se.nodes[3] -= 1;
          se.volume_id = volume_id; // CBM
	  selements.push(se);
	}
	else {
	  fprintf(stderr, "Error: element %d has a wrong type (%d)\n", num, t);
	  exit(1);
	}
      }
      else if (type_read == 3) {  // PJSA: node/dof dirichlet boundary conditions
        SBCond sdbc;
        sscanf(line, "%d %d %lg", &sdbc.node, &sdbc.dof, &sdbc.val);
        sdbc.node--; 
        sdbc.dof--;
        sdbcs.push(sdbc);
      }
    }
    else {
      fprintf(stdout, "    Found %s (%d)\n", key2, type_boco);
    }
  }

  fclose(fp);

  numNode = snodes.size();
  numFace = sbfaces.size();
  numElem = selements.size();
  numDirichlet = sdbcs.size();

  node.resize(numNode); // this is causing free(): invalid pointer
  faces = new Face[numFace];
  elements.resize(numElem);
  dbc = new BCond[numDirichlet];

  int i = 0;
  while (snodes.size()) {
    SNode sn = snodes.top();
    node[numNode-i-1][0] = sn.x[0];
    node[numNode-i-1][1] = sn.x[1];
    node[numNode-i-1][2] = sn.x[2];
    snodes.pop();
    ++i;
  }

  i = 0;
  while (sbfaces.size()) {
    SBFace sbf = sbfaces.top();
    faces[numFace-i-1][0] = sbf.nodes[0];
    faces[numFace-i-1][1] = sbf.nodes[1];
    faces[numFace-i-1][2] = sbf.nodes[2];
    faces[numFace-i-1].setBC(sbf.type);
    faces[numFace-i-1].setSurfaceID(sbf.surface_id);
    sbfaces.pop();
    ++i;
  }

  i = 0;
  while (selements.size()) {
    SElement se = selements.top();
    elements[numElem-i-1] = new Tetrahedron3D(se.nodes[0], se.nodes[1], se.nodes[2], se.nodes[3]);
    elements[numElem-i-1]->setVolumeID(se.volume_id);
    selements.pop();
    ++i;
  }

  i = 0;
  while (sdbcs.size()) {
    SBCond sdbc = sdbcs.top();
    dbc[i].nnum = sdbc.node;
    dbc[i].dofnum = sdbc.dof;
    dbc[i].val = sdbc.val;
    sdbcs.pop();
    ++i;
  }

}

//------------------------------------------------------------------------------

int cmp_nocase(const char* s, const char* s2, int n)
{

  int size_s = strlen(s);
  int size_s2 = strlen(s2);

  int count = 0;
  while (size_s != 0 && size_s2 != 0) {
    if (tolower(s[count]) != tolower(s2[count]))
      return (tolower(s[count]) < tolower(s2[count])) ? -1 : 1;
    ++count;
    if (n > 0 && count == n)
      return 0;
    --size_s;
    --size_s2;
  }

  size_s = strlen(s);
  size_s2 = strlen(s2);

  return (size_s == size_s2) ? 0 : (size_s < size_s2) ? -1 : 1;

}

//------------------------------------------------------------------------------

void FluidDomain::getGeometryFieldView(const char* meshname)  
{

  FILE* fp = fopen(meshname, "r");
 if (!fp) {
    fprintf(stderr, "Error: could not open file \'%s\'\n", meshname);
    exit(1);
  }

  // read boundary condition names
  char line[MAXLINE], item[MAXLINE];
  while (fgets(line, MAXLINE, fp) != 0) {
    if (cmp_nocase(line, "Boundary Table", 14) == 0)
      break;
  }
  fgets(line, MAXLINE, fp);
  int num_bcs;
  sscanf(line, "%d", &num_bcs);
  int* bcs = new int[num_bcs];
  int i;
  for (i=0; i<num_bcs; ++i) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%*s %s", item);
    for (int j=MIN_TYPE_BC; j<=MAX_TYPE_BC; ++j) {
      if (cmp_nocase(item, type_name[j], strlen(type_name[j])) == 0) {
	bcs[i] = j;
	if (bcs[i] == -1)
	  bcs[i] = -2;
	else if (bcs[i] == 1)
	  bcs[i] = 2;
	break;
      }
    }
    fprintf(stdout, "    Found %s (%d) boundary code\n", item, bcs[i]);
  }

  // read nodes
  while (fgets(line, MAXLINE, fp) != 0) {
    if (cmp_nocase(line, "Nodes", 5) == 0)
      break;
  }
  fgets(line, MAXLINE, fp);
  sscanf(line, "%d", &numNode);
  fprintf(stdout, "    Found %d nodes\n", numNode);
  node.resize(numNode);
  for (i=0; i<numNode; ++i) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%lf %lf %lf", &node[i][0], &node[i][1], &node[i][2]);
  }
 
  // read boundary faces
  while (fgets(line, MAXLINE, fp) != 0) {
    if (cmp_nocase(line, "Boundary Faces", 14) == 0)
      break;
  }
  fgets(line, MAXLINE, fp);
  sscanf(line, "%d", &numFace);
  fprintf(stdout, "    Found %d boundary faces\n", numFace);
  faces = new Face[numFace];
  for (i=0; i<numFace; ++i) {
    fgets(line, MAXLINE, fp);
    int t, nn;
    sscanf(line, "%d %d %d %d %d", &t, &nn, &faces[i][0], &faces[i][1], &faces[i][2]);
    if (nn != 3) {
      fprintf(stderr, "*** Error: wrong number of nodes (%d) for face %d\n", nn, i+1);
      exit(1);
    }
    faces[i][0] -= 1;
    faces[i][1] -= 1;
    faces[i][2] -= 1;
    faces[i].setBC(bcs[t-1]);
  }

  // read tetrahedra
  while (fgets(line, MAXLINE, fp) != 0) {
    if (cmp_nocase(line, "Elements", 8) == 0)
      break;
  }
  stack<SElement> selements;
  i = 0;
  while (fgets(line, MAXLINE, fp) != 0) {
    if (cmp_nocase(line, "Variables", 9) == 0)
      break;
    int type;
    SElement se;
    sscanf(line, "%d %*s %d %d %d %d", &type, &se.nodes[1], &se.nodes[0], &se.nodes[2], &se.nodes[3]);
    if (type != 1) {
      fprintf(stderr, "*** Error: wrong type (%d) for element %d\n", type, i+1);
      exit(1);
    }
    se.nodes[0] -= 1;
    se.nodes[1] -= 1;
    se.nodes[2] -= 1;
    se.nodes[3] -= 1;
    selements.push(se);
  }

  if (fp)
    fclose(fp);

  numElem = selements.size();
  elements.resize(numElem);

  i = 0;
  while (selements.size()) {
    SElement se = selements.top();
    elements[numElem-i-1] = new Tetrahedron3D(se.nodes[0], se.nodes[1], se.nodes[2], se.nodes[3]);
    selements.pop();
    ++i;
  }

  if (bcs) delete [] bcs;

}

//--------------------------------------------------------------------

void FluidDomain::getGeometrySinus(const char *file)  
{

  char line[MAXLINE];

  FILE *fp = fopen(file, "r");

  if (!fp) {
    fprintf(stderr, "*** ERROR: could not open file %s\n", file);
    exit(1);
  }

  fgets(line, MAXLINE, fp);
  sscanf(line, "%d %d %d", &numNode, &numElem, &numFace);

  node.resize(numNode);
  elements.resize(numElem);
  faces = new Face[numFace];

  int i;
  for (i = 0; i < numNode; i++) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%lf %lf %lf", &node[i][0], &node[i][1], &node[i][2]);    
  }

  int tets[4];
  for (i = 0; i < numElem; i++) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%d %d %d %d", tets+0, tets+1, tets+2, tets+3);
    elements[i] = new Tetrahedron3D(tets[0]-1, tets[1]-1, tets[2]-1, tets[3]-1);
  }

  int bc;
  for (i = 0; i < numFace; i++) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%d", &bc);
    faces[i].setBC(bc);
  }

  for (i = 0; i < numFace; i++) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%d %d %d", &faces[i][0], &faces[i][1], &faces[i][2]);    
    faces[i][0] -= 1;
    faces[i][1] -= 1;
    faces[i][2] -= 1;
  }

  fclose(fp);

}

//--------------------------------------------------------------------
// TDL:
// This function is currently obsolete.  It was written to support the
// redecomposition of a set of pre-sowered files.  The -redec 
// cmd is no longer supported, so this function no longer has a use.
Connectivity *FluidDomain::getBinaryGeometry(const char *prefix) 
{

  char name[MAXLINE];

  BinFileHandler::OffType start, locs[3];

  clusterNodeMap = new int*[numClusters];
  gl2ClNodeMap = new MapVec[numClusters];
  gl2ClElemMap = new MapVec[numClusters];
  gl2ClFaceMap = new MapVec[numClusters];

  int *numClusNodes = new int [numClusters];
  int *numClusElements = new int [numClusters];
  int *numClusFaces = new int [numClusters];

  int iClus;
  for (iClus=0; iClus<numClusters; ++iClus) {

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", prefix, suffix);
    BinFileHandler file(name, "rb");

    file.read(locs, 3);

    file.seek(locs[0]);
    file.read(numClusNodes+iClus, 1);
    file.seek(locs[1]);
    file.read(numClusElements+iClus, 1);
    file.seek(locs[2]);
    file.read(numClusFaces+iClus, 1);

  }

  numNode = 0;
  numElem = 0;
  numFace = 0;

  for (iClus=0; iClus<numClusters; ++iClus) {
    numNode += numClusNodes[iClus];
    numElem += numClusElements[iClus];
    numFace += numClusFaces[iClus];
  }

  node.resize(numNode);
  elements.resize(numElem);
  faces = new Face[numFace];

  int *ptrs = new int[numElem + 1];
  int *targets = new int[numElem];
  for (int i=0; i<numElem+1; ++i)
    ptrs[i] = i;

  numNode = 0;

  for (iClus = 0; iClus < numClusters; iClus++) {

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", prefix, suffix);
    BinFileHandler file(name, "rb");

    file.read(locs, 3);

    file.seek(locs[0] + sizeof(int));
    file.read(&start, 1);
    file.seek(start);

    clusterNodeMap[iClus] = new int[ numClusNodes[iClus] ];

    for (int iNode = 0; iNode < numClusNodes[iClus]; iNode++) {

      int glNode;
      file.read(&glNode, 1);

      clusterNodeMap[iClus][iNode] = glNode;
      gl2ClNodeMap[iClus][glNode] = iNode;

      double nd[3];
      file.read(nd, 3);
      for (int i = 0; i < 3; i++)
        node[clusterNodeMap[iClus][iNode]][i] = nd[i];

      if (clusterNodeMap[iClus][iNode] > numNode) 
	numNode = clusterNodeMap[iClus][iNode];

    }

    file.seek(locs[1] + sizeof(int));
    file.read(&start, 1);
    file.seek(start);

    for (int iElem = 0; iElem < numClusElements[iClus]; iElem++) {
      int tet[6];
      file.read(tet, 6);
      gl2ClElemMap[iClus][ tet[1] ] = iElem;

      tet[2] = clusterNodeMap[iClus][ tet[2] ];
      tet[3] = clusterNodeMap[iClus][ tet[3] ];
      tet[4] = clusterNodeMap[iClus][ tet[4] ];
      tet[5] = clusterNodeMap[iClus][ tet[5] ];

      elements[ tet[1] ] = new Tetrahedron3D(tet[2], tet[3], tet[4], tet[5]);
      targets[ tet[1] ] = iClus;
    }

    file.seek(locs[2] + sizeof(int));
    file.read(&start, 1);
    file.seek(start);

    for (int iFace = 0; iFace < numClusFaces[iClus]; iFace++) {
      int face[6];
      file.read(face, 6);
      gl2ClFaceMap[iClus][face[1]] = iFace; 
      faces[ face[1] ][0] = clusterNodeMap[iClus][ face[2] ];
      faces[ face[1] ][1] = clusterNodeMap[iClus][ face[3] ];
      faces[ face[1] ][2] = clusterNodeMap[iClus][ face[4] ];
      faces[ face[1] ].setBC(face[5]);
    }

  }

  ++numNode;

  if (numClusNodes) delete [] numClusNodes;
  if (numClusElements) delete [] numClusElements;
  if (numClusFaces) delete [] numClusFaces;

  Connectivity *elemToClus = new Connectivity(numElem, ptrs, targets);

  return elemToClus;

}

//-----------------------------------------------------------------------

void FluidDomain::getMatchInfo(const char *name)  
{

  FILE *fp = fopen(name, "r");

  if (!fp)  {
    fprintf(stderr, "*** ERROR: match file \'%s\' file does not exist\n", name);
    exit(-1);
  }

  char line[MAXLINE];
  fgets(line, MAXLINE, fp);
  fgets(line, MAXLINE, fp);
  sscanf(line, "%*s %d", &numMatchPts);

  matchPts = new int[numMatchPts];
  gapVec = new double[numMatchPts][3];

  // read match points
  int i;
  for (i = 0; i < numMatchPts; i++) {
    fgets(line, MAXLINE, fp);
    sscanf(line, "%d", matchPts+i);
    matchPts[i]--;
  }

  // read gap vectors
  i = 0;
  while (fgets(line, MAXLINE, fp) != 0) {
    sscanf(line, "%lf %lf %lf", gapVec[i], gapVec[i]+1, gapVec[i]+2);
    i++;
  }

  // check that # of gap vectors match number of matched points
  if (i == 0) {
    fprintf(stderr, " *** WARNING: setting all gap vectors to zero\n");
    for (i = 0; i < numMatchPts; i++) {
      gapVec[i][0] = 0.0;
      gapVec[i][1] = 0.0;
      gapVec[i][2] = 0.0;
    }
  }
  else if (i != numMatchPts)  {
    fprintf(stderr, " *** ERROR: incorrect number of gap vectors\n");
    exit(-1);
  }

  fclose(fp);

} 

//-----------------------------------------------------------------------

void FluidDomain::getBinaryMatchInfo(const char *prefix)  
{

  char name[MAXLINE];

  BinFileHandler::OffType start, loc;

  gl2ClMatchMap = new MapVec[numClusters];

  int *numClusNodes = new int [numClusters];

  int iClus;
  for (iClus=0; iClus<numClusters; ++iClus) {

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", prefix, suffix);
    BinFileHandler file(name, "rb");

    file.read(&loc, 1);

    file.seek(loc);
    file.read(numClusNodes+iClus, 1);

  }

  numMatchPts = 0;

  for (iClus=0; iClus<numClusters; ++iClus)
    numMatchPts += numClusNodes[iClus];

  matchPts = new int[numMatchPts];
  gapVec = new double[numMatchPts][3];

  numMatchPts = 0;

  for (iClus = 0; iClus < numClusters; iClus++) {

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", prefix, suffix);
    BinFileHandler file(name, "rb");

    file.read(&loc, 1);

    file.seek(loc + sizeof(int));
    file.read(&start, 1);
    file.seek(start);

    for (int iNode = 0; iNode < numClusNodes[iClus]; iNode++) {

      int index[2];
      file.read(index, 2);
      gl2ClMatchMap[iClus][ index[1] ] = iNode;
      matchPts[ index[1] ] = clusterNodeMap[iClus][ index[0] ];

      double nd[3];
      file.read(nd, 3);
      for (int i = 0; i < 3; i++)
        gapVec[ index[1] ][i] = nd[i];

      if (index[1] > numMatchPts) 
	numMatchPts = index[1];

    }

  }

  ++numMatchPts;

  if (numClusNodes) delete [] numClusNodes;

}

//-----------------------------------------------------------------------

void FluidDomain::makeConnectivity(Connectivity *elemToClus)  
{

  Domain::makeConnectivity(elemToClus);

  decomp->makeSubToFace(numFace, numElem, faces);

  if (numMatchPts)
    decomp->makeSubToMatch(decomp->getSubToNode(), numMatchPts, matchPts);

}

//---------------------------------------------------------------------

void FluidDomain::renumberNodes()  
{

  Domain::renumberNodes();

  // renumber faces nodes

  Connectivity *clusToSub = decomp->getClusToSub();
  Connectivity *subToFace = decomp->getSubToFace();

  for (int iClus = 0; iClus < numClusters; iClus++) {

    for (int iSub = 0; iSub < clusToSub->num(iClus); iSub++) {

      int glSub = (*clusToSub)[iClus][iSub];
      
      for (int iFace = 0; iFace < subToFace->num(glSub); iFace++) {
	int faceI = (*subToFace)[glSub][iFace];
	faces[faceI].renumberNodes(gl2ClNodeMap[iClus]);
      }

    }
    
  }

  // PJSA
  initBCData();

  for (int iClus = 0; iClus < numClusters; iClus++)  {
    renumberBC(gl2ClNodeMap[iClus], iClus);
    if(numDbcPerClus[iClus])
      fprintf(stderr," ... found %4d mesh motion Dirichlet boundary conditions in cluster %2d\n", numDbcPerClus[iClus], iClus);
  }


}

//---------------------------------------------------------------------

void FluidDomain::printGeometry(const char *outPrefix)  
{

  char outFileGeo[MAXLINE];
  char *suffix;

  Point3D *nodeSet = node.yield();
  Element **elems = elements.yield();

  Connectivity *clusToNode = decomp->getClusToNode();
  Connectivity *subToFace = decomp->getSubToFace();

  gl2ClElemMap = new MapVec[numClusters];
  gl2ClFaceMap = new MapVec[numClusters];

  for (int iCluster=0; iCluster<numClusters; iCluster++) {

    suffix = computeClusterSuffix(iCluster+1, numClusters);
    sprintf(outFileGeo, "%s.msh%s", outPrefix, suffix);
    BinFileHandler outputGeo(outFileGeo, "wb", FILE_VER);

    fprintf(stdout, "... Writing \'%s\'\n", outFileGeo);

    // save spot for ptr to TOC of nodes, elems and faces
    BinFileHandler::OffType firstline = outputGeo.tell();
    BinFileHandler::OffType *tocOffSets = new BinFileHandler::OffType[3];
    outputGeo.write(tocOffSets, 3);
    // PJSA: save spot for start of BCs
    outputGeo.write(&firstline, 1);

    // print out nodes
    BinFileHandler::OffType nodeListLoc = outputGeo.tell();
    printNodes(iCluster, outputGeo, nodeSet);

    // print out elements
    BinFileHandler::OffType elemListLoc = outputGeo.tell();
    int numClusElems;
    int *elemLoc = decomp->printElements(iCluster, outputGeo, numClusElems,
					 elems, gl2ClElemMap[iCluster]);

    // print out faces
    BinFileHandler::OffType faceListLoc = outputGeo.tell();
    int numClusFaces;
    int *faceLoc = printFaces(iCluster, outputGeo, numClusFaces, subToFace,
			      gl2ClFaceMap[iCluster]);
    
    // PJSA: print out dirichlet boundary conditions
    BinFileHandler::OffType startOfBCsLoc = outputGeo.tell();
    printBC(outputGeo, iCluster);

    // print table of contents for nodes
    tocOffSets[0] = outputGeo.tell();
    int numClusNodes = clusToNode->num(iCluster);
    outputGeo.write(&numClusNodes, 1);
    outputGeo.write(&nodeListLoc, 1);

    // print out table of contents for elements
    tocOffSets[1] = outputGeo.tell();
    outputGeo.write(&numClusElems, 1);
    outputGeo.write(&elemListLoc, 1);
    outputGeo.write(elemLoc, numClusElems);
    if (elemLoc) delete [] elemLoc;

    // print out table of contents for faces
    tocOffSets[2] = outputGeo.tell();
    outputGeo.write(&numClusFaces, 1);
    outputGeo.write(&faceListLoc, 1);
    outputGeo.write(faceLoc, numClusFaces);
    if (faceLoc) delete [] faceLoc;

    // go back to beginning of file and write location of TOC
    outputGeo.seek(firstline);
    outputGeo.write(tocOffSets, 3);
    // PJSA: write loc of start of BCs
    outputGeo.write(&startOfBCsLoc, 1);

  }

}

//---------------------------------------------------------------------

void FluidDomain::printGeometryEnsight(const char *outPrefix)  
{

  char outFileGeo[256];
  char outFileMat[256];
  char *suffix;

  Point3D *nodeSet = node.yield();
  Element **elems  = elements.yield();

  Connectivity *clusToNode = decomp->getClusToNode();
  Connectivity *subToFace = decomp->getSubToFace();

  gl2ClElemMap = new MapVec[numClusters];
  gl2ClFaceMap = new MapVec[numClusters];

  sprintf(outFileGeo, "%s.geo", outPrefix);    
  sprintf(outFileMat, "%s.mat", outPrefix);    
  fprintf(stdout, "... Writing Ensight geometry file \'%s\'\n", outFileGeo);

  // open file
  EnsightFileHandler outputGeo(outFileGeo, "wb");
  EnsightFileHandler outputMat(outFileMat, "wb");

  // write header
  outputGeo.header("ENSIGHT GOLD GEOMETRY file",
		   "Written by Sower");
  outputMat.header("ENSIGHT GOLD MATERIAL file");
  
  for (int iCluster=0; iCluster<numClusters; iCluster++) {
    // start new part
    char part_name[80];
    sprintf(part_name, "Fluid mesh, part %d", iCluster);
    outputGeo.part(part_name);
    outputMat.part();

    // print nodes for this part
    printNodes(iCluster, outputGeo, nodeSet);

    // print elements for this part
    int numClusElems;
    decomp->printElements(iCluster, outputGeo, outputMat, numClusElems,
			  elems, gl2ClElemMap[iCluster]);
  }

}

//---------------------------------------------------------------------

int *FluidDomain::printFaces(int iClus, BinFileHandler &outFileGeo, int &numClusFaces, 
			     Connectivity *subToFace, MapVec &gl2ClFaces)  
{

  // get connectivities for looping
  Connectivity *clusToSub = decomp->getClusToSub();
  int numSubs = clusToSub->num(iClus);

  // count total # of faces in cluster
  numClusFaces = 0;

  int iSub;
  for (iSub = 0; iSub < numSubs; iSub++)
    numClusFaces += subToFace->num((*clusToSub)[iClus][iSub]);

  int *facePtrs = new int[numClusFaces];

  int count = 0;
  int numInts = 0;  // # of integers required for face data
  
  for (iSub = 0; iSub < numSubs; iSub++) {

    // get subdomains in current cluster
    int subD = (*clusToSub)[iClus][iSub];

    for (int iFace = 0; iFace < subToFace->num(subD); iFace++) {

      // get current face in subdomain
      int faceI = (*subToFace)[subD][iFace];
      gl2ClFaces[faceI] = count;

      int faceNodes[7];
      faceNodes[0] = faces[faceI].getType();
      faceNodes[1] = faceI;
      faceNodes[2] = faces[faceI][0];
      faceNodes[3] = faces[faceI][1];
      faceNodes[4] = faces[faceI][2];
      faceNodes[5] = faces[faceI].getBC();
      faceNodes[6] = faces[faceI].getSurfaceID();

      outFileGeo.write(faceNodes, 7);

      facePtrs[count] = numInts;
      numInts += 7;
      count++;
    }

  }
  
  return facePtrs;

}

//---------------------------------------------------------------------

void FluidDomain::printMatchInfo(const char *outPrefix)  
{

  char outFileMatch[MAXLINE];

  Connectivity *clusToSub = decomp->getClusToSub();
  Connectivity *clusToMatchPts = decomp->getClusToMatch();

  gl2ClMatchMap = new MapVec[numClusters];

  for (int iCluster = 0; iCluster < numClusters; iCluster++) {

    char *suffix = computeClusterSuffix(iCluster+1, numClusters);
    sprintf(outFileMatch, "%s.match%s", outPrefix, suffix);
    BinFileHandler outputMatch(outFileMatch, "wb", FILE_VER);

    fprintf(stdout, "... Writing \'%s\'\n", outFileMatch);

    BinFileHandler::OffType firstline = outputMatch.tell();
    outputMatch.write(&firstline, 1);
    BinFileHandler::OffType startOfData = outputMatch.tell();
    
    int numClusMatchPts = clusToMatchPts->num(iCluster);
    for (int i=0; i<numClusMatchPts; ++i) {
      int glmatch = (*clusToMatchPts)[iCluster][i];
      int glnode = matchPts[glmatch];
      gl2ClMatchMap[iCluster][glmatch] = i;

      int matchData[2] = {gl2ClNodeMap[iCluster][glnode], glmatch};

      outputMatch.write(matchData, 2);
      outputMatch.write(gapVec[glmatch], 3);
    }

    BinFileHandler::OffType tocLoc = outputMatch.tell();
    outputMatch.write(&numClusMatchPts, 1);
    outputMatch.write(&startOfData, 1);
    outputMatch.seek(firstline);
    outputMatch.write(&tocLoc, 1);

  }

}

//---------------------------------------------------------------------

void FluidDomain::printDecomp(const char *outPrefix) 
{

  char outFileDec[MAXLINE];

  Connectivity *clusToSub = decomp->getClusToSub();

  int *numSubNeighbors = decomp->getNumNeighbors();
  int **subNeighbors = decomp->getSubNeighbors();
  Connectivity **interfNodes = decomp->getInterfNodes();

  for (int iCluster=0; iCluster < numClusters; iCluster++) {

    char *suffix = computeClusterSuffix(iCluster+1, numClusters);
    sprintf(outFileDec, "%s.dec%s", outPrefix, suffix);
    BinFileHandler outputDec(outFileDec, "w", FILE_VER);

    fprintf(stdout, "... Writing \'%s\'\n", outFileDec);

    int numClusSubs = clusToSub->num(iCluster);

    outputDec.write(&numClusSubs, 1);

    BinFileHandler::OffType offsetLoc = outputDec.tell();

    BinFileHandler::OffType *subOffSets = new BinFileHandler::OffType[numClusSubs];

    outputDec.write(subOffSets, numClusSubs);

    for (int iSub = 0; iSub < numClusSubs; iSub++) {
      
      subOffSets[iSub] = outputDec.tell();

      int glSub = (*clusToSub)[iCluster][iSub];

      printRangesInfo(outputDec, glSub, decomp->getSubToNode(), gl2ClNodeMap[iCluster]);

      printRangesInfo(outputDec, glSub, decomp->getSubToElem(), gl2ClElemMap[iCluster]);

      printRangesInfo(outputDec, glSub, decomp->getSubToFace(), gl2ClFaceMap[iCluster]);

      printInterfaceInfo(outputDec, numSubNeighbors[glSub], subNeighbors[glSub], 
			 interfNodes[glSub], gl2ClNodeMap[iCluster]);

      if (numMatchPts) 
	printRangesInfo(outputDec, glSub, decomp->getSubToMatch(), gl2ClMatchMap[iCluster]);

    }

    outputDec.seek(offsetLoc);
    outputDec.write(subOffSets, numClusSubs);

  }

}

