#include "Main.h"
#include "Domain.h"
#include "Element.h"
#include <Postpro.h>

extern void merge(FileNames &, PostOpts &);
extern void mergeStructData(FileNames &, int = -1);
extern void split(FileNames &, PostOpts &);
extern int readCommandLine(int, char **, int &, int &, int *,
                           FileNames &, PostOpts &, int &);

//-------------------------------------------------------------------------------

int main(int argc, char** argv) 
{

  int numNumCpus, numClusters, numCpus[100];
  FileNames fileNames;
  PostOpts postOpts;

  int extractNum = -1;
  int type = readCommandLine(argc, argv, numClusters, numNumCpus, numCpus,
                             fileNames, postOpts, extractNum);

  Domain *domain;

  if (type == 1) {
    merge(fileNames, postOpts);
    exit(0);
  }
  else if (type == 2) {
    split(fileNames, postOpts);
    exit(0);
  }
  else if (type == 3) {
    fprintf(stdout, "... Sowering a fluid domain\n");
    domain = new FluidDomain(numClusters);
  }
  else if (type == 6)  {
    mergeStructData(fileNames);
    exit(0);
  }
  else if (type == 7)  {
    mergeStructData(fileNames, extractNum);
    exit(0);
  }
  else {
    fprintf(stdout, "... Sowering a structure domain\n");
    domain = new StructureDomain(numClusters);
  }

  const char *meshfile = fileNames.geo;
  const char *decompfile = fileNames.dec;
  const char *matchfile = fileNames.match;
  const char *outPrefix = fileNames.prefix;

  fprintf(stdout, "... Reading the mesh from \'%s\'\n", meshfile);

  Connectivity *elemToClus = 0;

  domain->getGeometry(meshfile);

  if (matchfile[0] != 0) {
    fprintf(stdout, "... Reading the matched info from \'%s\'\n", matchfile);
    domain->getMatchInfo(matchfile);
  }

  fprintf(stdout, "... Reading the decomposition from \'%s\'\n", decompfile);

  domain->getDecomp(decompfile, numNumCpus, numCpus);
 
  fprintf(stderr, "... Building the connectivities\n");

  domain->makeConnectivity(elemToClus);


  fprintf(stdout, "... Renumbering the nodes per cluster\n");

  domain->createClusterToGlobalNodeMap();

  fprintf(stdout, "... renumber nodes\n");
  domain->renumberNodes();

  if (fileNames.ensight)
    domain->printGeometryEnsight(outPrefix);
  else {
    domain->printGeometry(outPrefix);

    if (matchfile[0] != 0)
      domain->printMatchInfo(outPrefix);

    domain->printDecomp(outPrefix);
    
    domain->printGlobalConnectivity(outPrefix);

    for (int k=0; k<numNumCpus; ++k)
      if (numCpus[k] > 0)
        domain->printCpuToSubMap(numCpus[k], outPrefix, type);
  }

}

//---------------------------------------------------------------------------

int readCommandLine(int argc, char **argv, int &numClusters, 
                    int &numNumCpus, int *numCpus,
                    FileNames &fn, PostOpts &po, int &extractNum)  {

  int i = 0;
  int t0 = 0;
  int t1 = 0;
  int t2 = 0;
  int t3 = 0;
  int type = -1;
  
  int numBCs = 0;
  int bcs[100];
  bool isBcs = false;
  bool isSplit = false;

  numClusters = 1;
  numNumCpus = 0;
  int opType = 0;    // specifies operation type
                     // 0 for create bin files, 1 for merge, 2 for split
                     // 3 for single node/element extraction

  int meshType = -1;  // determine mesh type: 1 for fluid, 2 for structure

  do {

    if (strncmp(argv[i], "-ascii", 2) == 0) { fn.ascii = true; ++t3; }
    if (strncmp(argv[i], "-binary", 3) == 0) { fn.binary = true; }
    if (strncmp(argv[i], "-bc", 3) == 0) {
      isBcs = true;
      bcs[numBCs++] = atoi(argv[i+1]);
    }

    if (strncmp(argv[i], "-cluster", 3) == 0) 
      numClusters = atoi(argv[i+1]);

    if (strncmp(argv[i], "-con", 3) == 0) { fn.global = argv[i+1]; ++t1; ++t2; ++t3; }
    if (strncmp(argv[i], "-cpu", 3) == 0) 
      numCpus[numNumCpus++] = atoi(argv[i+1]);

    if (strncmp(argv[i], "-dec", 4) == 0) { fn.dec = argv[i+1]; }
    if (strncmp(argv[i], "-density", 4) == 0) { fn.dens = argv[i+1]; }
    if (strncmp(argv[i], "-displacements", 3) == 0) { fn.disp = argv[i+1]; }
    if (strncmp(argv[i], "-elements", 3) == 0) { fn.elements = argv[i+1]; }
    if (strncmp(argv[i], "-ensight", 3) == 0) { fn.ensight = true; }
    if (strncmp(argv[i], "-extract", 3) == 0) { 
      opType = 3; 
      extractNum = atoi(argv[i+1]); 
    }
    if (strncmp(argv[i], "-eps", 3) == 0) { fn.eps = argv[i+1]; }
    if (strncmp(argv[i], "-fluid", 3) == 0) { meshType = 1; }
    if (strncmp(argv[i], "-freq", 3) == 0) { po.freqResPicks = atoi(argv[i+1]); }
    if (strncmp(argv[i], "-inlet_mach", 2) == 0) { po.machInf = atof(argv[i+1]); }
    if (strncmp(argv[i], "-k", 2) == 0) { fn.k = argv[i+1]; }
    if (strncmp(argv[i], "-range", 3) == 0) { 
      po.resPicks[po.numResPicks][0] = atoi(argv[i+1]);
      po.resPicks[po.numResPicks][1] = atoi(argv[i+2]);
      ++po.numResPicks; 
    }
    if (strncmp(argv[i], "-load", 3) == 0) { fn.load = argv[i+1]; }
    if (strncmp(argv[i], "-mach", 4) == 0) { fn.mach = argv[i+1]; }
    if (strncmp(argv[i], "-match", 4) == 0) { fn.match = argv[i+1]; }
    if (strncmp(argv[i], "-merge", 4) == 0) { opType = 1; }
    if (strncmp(argv[i], "-mesh", 4) == 0) { fn.geo = argv[i+1]; ++t0; ++t1; ++t2; ++t3; }
    if (strncmp(argv[i], "-name", 3) == 0) { fn.name = argv[i+1]; }
    if (strncmp(argv[i], "-nodes", 3) == 0) { fn.nodes = argv[i+1]; }
    if (strncmp(argv[i], "-output", 2) == 0) { fn.prefix = argv[i+1];  ++t2; }
    if (strncmp(argv[i], "-pressure", 2) == 0) { fn.pres = argv[i+1]; }
    if (strncmp(argv[i], "-result", 4) == 0) { fn.result = argv[i+1]; ++t2; ++ t3;}
    if (strncmp(argv[i], "-skin", 3) == 0) { fn.skin = argv[i+1]; }
    if (strncmp(argv[i], "-nodemap", 6) == 0) { fn.nodemap = argv[i+1]; }
    if (strncmp(argv[i], "-split", 3) == 0) { 
      opType = 2;
      isSplit = true; 
    }
    if (strncmp(argv[i], "-struct", 3) == 0) { meshType = 2; }
    if (strncmp(argv[i], "-temperature", 2) == 0) { fn.temp = argv[i+1]; }
    if (strncmp(argv[i], "-velocities", 2) == 0) { fn.vel = argv[i+1]; }

    ++i;

  } while (i < argc);

  if(isBcs && isSplit) { fn.isSdesign = true; }

  switch (meshType)  {

    case 1: // Fluid Mesh
      switch (opType)  {

        case 0:
          type = 3;  // create bin input files for fluid
          break;
        case 1:
          type = 1;  // merge binary fluid solutions
          break;
        case 2:
          type = 2;  // split fluid solution
          break;
      }
      break;

    case 2:   // Structure Mesh
      switch (opType)  {

        case 0:
          type = 5;     // create bin input files for structure
          break;
        case 1:
          type = 6;     // merge binary structure solutions
          break;
        case 3:
          type = 7;     // data extraction of 1 node or element
          break;
        default:    // error msg.
          fprintf(stderr," ERROR: Split Operation not supported for structures: %d\n", type);
          exit(1);
          break;
      }


      break;
 
    default:  // error msg.

      fprintf(stdout, "Incorrect usage\n");
      fprintf(stdout, "Usage: sower -fluid [-merge] [-split]\n");
      fprintf(stdout, "       sower -struct [-merge] [-extract]\n");
      exit(1);
      break;
  
  }

  if (type <= 0) {
    fprintf(stdout, "Incorrect usage\n");
    fprintf(stdout, "Usage: sower -fluid | -struct | -merge | -split\n");
    exit(1);
  }
  else if ((type == 3 || type == 5) && t0 != 1) {
    fprintf(stdout, "Incorrect usage\n");
    if (meshType == 1)
      fprintf(stdout, "Usage: sower -fluid\n");
    else if (meshType == 2)
      fprintf(stdout, "Usage: sower -struct\n");
    fprintf(stdout, "             -mesh <mesh file> [-match <matcher file>]\n");
    fprintf(stdout, "             [-dec <decomposition file>] [-cpu <number cpus>]\n");
    fprintf(stdout, "             [-cluster <number clusters>] [-output <output file prefix>]\n");
    exit(1);
  }
  else if (type == 1 && t1 != 2) {
    fprintf(stdout, "Incorrect usage\n");
    fprintf(stdout, "Usage: sower -fluid -merge\n");
    fprintf(stdout, "             -con <connectivity file>\n");
    fprintf(stdout, "             -mesh <mesh file prefix>\n");
    fprintf(stdout, "             [-result <binary result file prefix>]\n");
    fprintf(stdout, "             [-output <ascii output file prefix>] [-binary]\n");
/*
    fprintf(stdout, "             [-density <variable name>]\n");
    fprintf(stdout, "             [-pressure <variable name>]\n");
    fprintf(stdout, "             [-temperature <variable name>]\n");
    fprintf(stdout, "             [-k <variable name>]\n");
    fprintf(stdout, "             [-eps <variable name>]\n");
    fprintf(stdout, "             [-mach <variable name>]\n");
    fprintf(stdout, "             [-velocities <variable name>]\n");
    fprintf(stdout, "             [-displacements <variable name>]\n");
*/
    fprintf(stdout, "             [-frequency <output frequency>] [-range <res# res#>]\n");
    fprintf(stdout, "             [-nodes <node set name>] [-elements <element set name>]\n");
    fprintf(stdout, "             [-name <result name>] [-load <load case name>]\n");
    fprintf(stdout, "             [-skin <output skin name>] [-bc <boundary code>]\n");
    exit(1);
  }
  else if (type == 2 && t2 != 4 && !(fn.isSdesign)) {
    fprintf(stdout, "Incorrect usage\n");
    fprintf(stdout, "Usage: sower -fluid -split\n");
    fprintf(stdout, "             -con <connectivity file>\n");
    fprintf(stdout, "             -mesh <mesh file prefix>\n");
    fprintf(stdout, "             -result <binary result file> [-ascii]\n");
    fprintf(stdout, "             -output <binary output file prefix>\n");
    exit(1);
  }
  else if (type == 2 && t3 !=4 && fn.isSdesign) {
    fprintf(stdout, "Incorrect usage\n");
    fprintf(stdout, "Usage: sower -fluid -split\n");
    fprintf(stdout, "             -con <connectivity file>\n");
    fprintf(stdout, "             -mesh <mesh file prefix>\n");
    fprintf(stdout, "             -result <binary result file> -ascii\n");
    fprintf(stdout, "             -bc <boundary code>\n");
    exit(1);
  }
  else if(type == 6 && t2 != 3)  {
    fprintf(stdout, "Incorrect usage\n");
    fprintf(stdout, "Usage: sower -struct -merge\n");
    fprintf(stdout, "             -con <connectivity file>\n");
    fprintf(stdout, "             -mesh <mesh file prefix>\n");
    fprintf(stdout, "             -result <binary result file>\n");
    exit (1);
  }

  if (type == 3 || type == 5 || (type == 2 && t3 == 4)) {
    if (fn.prefix[0] == 0) 
      fn.prefix = "OUTPUT";
  }

  if (numBCs > 0) {
    po.numBCs = numBCs;
    for (i=0; i<numBCs; ++i)
      po.bcs[i] = bcs[i];
  }

  if (fn.name[0] == 0) {
    int i;
    for(i = strlen(fn.prefix); i-- > 0; )
      if(fn.prefix[i] == '/')
    break;
    fn.name = fn.prefix+i+1;
  }

  if (fn.nodes[0] == 0)  {
    if (fn.skin[0] != 0)
      fn.nodes = "SkinNodes";
    else if (meshType == 1)
      fn.nodes = "FluidNodes";
    else
      fn.nodes = "StructureNodes";
  }
  
  if (fn.elements[0] == 0)
    if (fn.skin[0] != 0)
      fn.elements = "SkinMesh";
    else
      fn.elements = "FluidMesh";
  
  return type;

}

//-------------------------------------------------------------------------------
