#include <Postpro.h>

#include <VarFcnDesc.h>
#include <Connectivity.h>
#include <Vector.h>
#include <BinFileHandler.h>
#include <EnsightFileHandler.h>
#include <iostream>

#ifdef BINARY_XPOST
#include <BinXpost.h>
#endif

#include <cmath>
#include <alloca.h>

PostOp *postOpObj = new FluidPostOp;

//-------------------------------------------------------------------------------

#ifdef BINARY_XPOST
void writeXpostHeader(const char *prefix, const char *type, 
                      const char *load, const char *nodes, 
                      int numNodes, XpostSet *&fp, const char *sol = 0)
{

  if (prefix[0] == 0) return;

  const char* namesol = prefix;
  if (sol)
    namesol = sol;

  char name[MAXLINE];
  sprintf(name, "%s.xpost.bin", prefix);
  if (strcmp(type, "Scalar") == 0)
    fp = new Scalarset(name, namesol, load, nodes, numNodes);
  else
    fp = new Vectorset(name, namesol, load, nodes, numNodes);

  fp->writeHeader();

}
#else
void writeXpostHeader(const char *prefix, const char *type, 
                      const char *load, const char *nodes, 
                      int numNodes, FILE *&fp, const char *sol = 0)
{

  if (prefix[0] == 0) return;

  const char* namesol;
  if (sol)
    namesol = sol;
  else {
    int i;
    for(i = strlen(prefix); i-- > 0; )
      if(prefix[i] == '/') 
    break;
    namesol = prefix+i+1;
  }

  char name[MAXLINE];
  sprintf(name, "%s.xpost", prefix);
  fp = fopen(name, "w");

  fprintf(fp, "%s %s under %s for %s\n", type, namesol, load, nodes);
  fprintf(fp, "%d\n", numNodes);

}
#endif

//-------------------------------------------------------------------------------

void writeXpostSolution(int size, int dim, double *tag, int type, double *U, FILE *fp)
{
  int neq = 1;

  if (type == PostOp::ASCII)
    neq = dim;
  else if (type == PostOp::VEL || type == PostOp::DISP)
    neq = 3;

  if (tag)
    fprintf(fp, "%g\n", *tag);

  for (int i=0; i<size; ++i) {
    fprintf(fp, "%e", U[i*neq + 0]);
    for (int j=1; j<neq; ++j)
      fprintf(fp, " %e", U[i*neq + j]);
    fprintf(fp, "\n");
  }

}

//-------------------------------------------------------------------------------
#ifdef FIELDVIEW

void writeFieldviewHeader(const char *prefix, int numNodes, double tag, FILE *&fp, const char *sol = 0, const char *ResId = 0)
{

  if (prefix[0] == 0) return;

  const char* namesol = prefix;
  if (sol)
    namesol = sol;

  char name[MAXLINE];
  sprintf(name, "%s%s.fvuns", prefix, ResId);
  fp = fopen(name, "w");

  fprintf(fp,"FIELDVIEW_Results 2 4\n");
  fprintf(fp,"Constants\n");
  fprintf(fp,"%f  %f  %f  %f\n",tag,0.0,0.0,0.0);
  fprintf(fp,"Grids 1\n");
  if (strcmp(namesol, "Velocity") == 0 || strcmp(namesol, "velocity") == 0 || strcmp(namesol, "VELOC") == 0) {
     fprintf(fp,"Variable Names 3\n");
     fprintf(fp,"uvel; velocity \n");
     fprintf(fp,"vvel \n");
     fprintf(fp,"wvel \n");
  } else {
     fprintf(fp,"Variable Names 1\n");
     fprintf(fp,"%s\n",namesol);
  }
  fprintf(fp,"Nodes  %d\n",numNodes);
  fprintf(fp,"Variables\n");

}

//-------------------------------------------------------------------------------

void writeFieldviewSolution(int size, int dim, int type, double *U, FILE *fp)
{

  int neq = 1;

  if (type == PostOp::ASCII)
    neq = dim;
  else if (type == PostOp::VEL || type == PostOp::DISP)
    neq = 3;

  for (int i=0; i<neq; ++i)
    for(int j=0; j<size; ++j)
      fprintf(fp,"%g\n",U[j*neq + i]);

}

//-------------------------------------------------------------------------------

// CBM: to avoid confusion with cluster suffix

char *computeFVresultSuffix(int num, int maxNum)
{

return computeClusterSuffix(num, maxNum);

}

#endif

//-------------------------------------------------------------------------------

void processSolution(int size, int dim, double tag, double machInf, double *U, 
                     double (*X)[3], int type, double *W)
{

  double *V = reinterpret_cast<double *>(alloca(dim * sizeof(double)));

  VarFcn *varFcn = 0;

  if (dim == 5)
    varFcn = new VarFcnEuler3D(1.4);
  else if (dim == 6)
    varFcn = new VarFcnSA3D(1.4);
  else if (dim == 7)
    varFcn = new VarFcnKE3D(1.4);

  for (int i=0; i<size; ++i) {

    switch (type) {
      
    case PostOp::ASCII : {
      for (int j=0; j<dim; ++j)
        W[i*dim + j] = U[i*dim + j];
      break;
    }
    case PostOp::DISP : {
      for (int j=0; j<3; ++j)
        W[i*3 + j] = U[i*3 + j] - X[i][j];
      break;
    }
    case PostOp::DENS :
    case PostOp::PRES :
    case PostOp::TEMP :
    case PostOp::KEN :
    case PostOp::EPS :
    case PostOp::MACH :
    case PostOp::VEL : {
      varFcn->conservativeToPrimitive(&U[i*dim], V);
      double rho = varFcn->getDensity(V);
      double p = varFcn->getPressure(V);
      double cp = varFcn->computeCp(V, machInf);
      double T = varFcn->computeTemperature(V);
      double k = varFcn->getTurbulentKineticEnergy(V);
      double eps = varFcn->getTurbulentDissipationRate(V);
      Vec3D u = varFcn->getVelocity(V);
      double mach = varFcn->computeMachNumber(V);
      /*
      double machinf = 0.2;
      double gam = varFcn->getGamma();
      double ppinf = gam * machinf*machinf * p;
      p = (1.0 + 0.5 * (gam-1.0) * mach*mach) / (1.0 + 0.5 * (gam-1.0) * machinf*machinf);
      p = ppinf * pow(p, gam / (gam - 1.0));
      */

      if (rho <= 0.0)
        fprintf(stderr, " *** ERROR: negative density (%e) for node %d (tag=%e)\n", 
                rho, i+1, tag);

      if (p <= 0.0)
        fprintf(stderr, " *** ERROR: negative pressure (%e) for node %d (tag=%e)\n", 
                p, i+1, tag);

      if (k < 0.0)
        fprintf(stderr, " *** ERROR: negative k (%e) for node %d (tag=%e)\n", 
                p, i+1, tag);

      if (eps < 0.0)
        fprintf(stderr, " *** ERROR: negative eps (%e) for node %d (tag=%e)\n", 
                p, i+1, tag);

      switch (type) {

      case PostOp::DENS : {
        W[i] = rho;
        break;
      }
      case PostOp::PRES : {
        if (machInf > 0)
          W[i] = cp;
        else
          W[i] = p;
        break;
      }
      case PostOp::TEMP : {
        W[i] = T;
        break;
      }
      case PostOp::KEN : {
        W[i] = k;
        break;
      }
      case PostOp::EPS : {
        W[i] = eps;
        break;
      }
      case PostOp::MACH : {
        W[i] = mach;
        break;
      }
      case PostOp::VEL : {
        for (int j=0; j<3; ++j)
          W[i*3 + j] = u[j];
        break;
      }
      }
    }
    }
  }
 
  if (varFcn) delete varFcn;
 
}

//----------------------------------------------------------------------

void merge(FileNames &fn, PostOpts &po)
{

  int i, j;
  FluidPostOp postop;

  // reading topology

  postop.initialize(fn.global, fn.geo);

  int numNodes = postop.getNumberOfNodes();
  int size = numNodes;

  int numSkinNodes;
  int *skinNodeMap;
  if (fn.skin[0] != 0) {
    numSkinNodes = postop.extractSkinSurface(fn, po, skinNodeMap);
    size = numSkinNodes;
  }
  else if (fn.skin[0] == 0 && fn.result[0] == 0)
    postop.writeTopology(fn);

  if (fn.result[0] == 0) return;

  // reading header from binary

  int header[3];
  postop.readSolutionHeader(fn.result, header, FluidPostOp::DISTRIBUTED);

  int dim = header[1];
  int numRes = header[2];

  if (numRes <= 0)  {
    fprintf(stderr, "ERROR: Bad Number of Input Results: %d\n", numRes);
    exit(-1);
  }

  int *activeRes;
  int numActiveRes = po.tagResults(numRes, activeRes);

  header[0] = size;
  header[2] = numActiveRes;  

  // writing headers for ascii and binary

#ifdef BINARY_XPOST
  XpostSet* fp[PostOp::SIZE];
#else
  FILE *fp[PostOp::SIZE];
#endif
  for (i=0; i<PostOp::SIZE; ++i)
    fp[i] = 0;

#ifdef SYMMETRY
  int save_size = size;
  size *= 2;
#endif

#ifndef FIELDVIEW
  if (fn.prefix[0] != 0 && fn.binary)
    postop.writeSolutionHeader(fn.prefix, header, FluidPostOp::SHARED);
  else {
    if (dim == 1)
      writeXpostHeader(fn.prefix, "Scalar", fn.load, fn.nodes, size, fp[PostOp::ASCII], fn.name);
    else
      writeXpostHeader(fn.prefix, "Vector", fn.load, fn.nodes, size, fp[PostOp::ASCII], fn.name);
    writeXpostHeader(fn.dens, "Scalar", fn.load, fn.nodes, size, fp[PostOp::DENS]);
    writeXpostHeader(fn.pres, "Scalar", fn.load, fn.nodes, size, fp[PostOp::PRES]);
    writeXpostHeader(fn.temp, "Scalar", fn.load, fn.nodes, size, fp[PostOp::TEMP]);
    writeXpostHeader(fn.k,    "Scalar", fn.load, fn.nodes, size, fp[PostOp::KEN]);
    writeXpostHeader(fn.eps,  "Scalar", fn.load, fn.nodes, size, fp[PostOp::EPS]);
    writeXpostHeader(fn.mach, "Scalar", fn.load, fn.nodes, size, fp[PostOp::MACH]);
    writeXpostHeader(fn.vel,  "Vector", fn.load, fn.nodes, size, fp[PostOp::VEL], "VELOC");
    writeXpostHeader(fn.disp, "Vector", fn.load, fn.nodes, size, fp[PostOp::DISP]);
  }
#endif

#ifdef SYMMETRY
  size = save_size;
#endif

  // writing solution for ascii and binary

  double *Ufull = new double[numNodes * dim];
  double (*Xfull)[3] = postop.getNodes();
  double *Uskin = 0;
  double (*Xskin)[3] = 0;
  double *Uxpost = new double[size * dim];

  if (fn.skin[0] != 0) {
    Uskin = new double[numSkinNodes * dim];
    Xskin = new double[numSkinNodes][3];
    for (i=0; i<numSkinNodes; ++i)
      for (j=0; j<3; ++j)
        Xskin[i][j] = Xfull[ skinNodeMap[i] ][j];    
  }

  int iActiveRes = 0;
  for (int iGlobRes=0; iGlobRes<numRes; ++iGlobRes) {

    if (activeRes[iGlobRes] == 0) continue;
    
    double tag;
    postop.readSolution(fn.result, iGlobRes, dim, tag, Ufull, FluidPostOp::DISTRIBUTED);

    fprintf(stdout, " ... Processing result %d with tag %g    \n", iGlobRes+1, tag);
    fflush(stdout);

    if (fn.ensight)
      postop.writeSolutionEnsight(fn.prefix, iActiveRes, dim, tag, Ufull, FluidPostOp::SHARED);
    else if (fn.prefix[0] != 0 && fn.binary)
      postop.writeSolution(fn.prefix, iActiveRes, dim, tag, Ufull, FluidPostOp::SHARED);
    else {
#ifdef FIELDVIEW
      char *fvSuffix;
      fvSuffix = computeFVresultSuffix(iGlobRes+1, numRes);
      writeFieldviewHeader(fn.prefix, size, tag, fp[PostOp::ASCII], fn.name, fvSuffix);
#endif
      double *U = Ufull;
      double (*X)[3] = Xfull;

      if (fn.skin[0] != 0) {
        for (i=0; i<numSkinNodes; ++i)
          for (j=0; j<dim; ++j)
            Uskin[i*dim + j] = Ufull[ skinNodeMap[i]*dim + j];
        U = Uskin;
        X = Xskin;
      }

      for (i=0; i<PostOp::SIZE; ++i) {
        if (fp[i]) {
          processSolution(size, dim, tag, po.machInf, U, X, i, Uxpost);
#ifdef BINARY_XPOST
          fp[i]->writeXpostSolution(tag, Uxpost);
#elif FIELDVIEW
          writeFieldviewSolution(size, dim, i, Uxpost, fp[i]);
#else
          writeXpostSolution(size, dim, &tag, i, Uxpost, fp[i]);
#ifdef SYMMETRY
          writeXpostSolution(size, dim, 0, i, Uxpost, fp[i]);
#endif
#endif
        }
      }
    }

    ++iActiveRes;
    
  }
  
  fprintf(stdout, "\n");

#ifdef BINARY_XPOST
  for (i=0; i<PostOp::SIZE; ++i)
    if (fp[i]) delete fp[i];
#else
  for (i=0; i<PostOp::SIZE; ++i)
    if (fp[i]) fclose(fp[i]);
#endif

  if (Ufull) delete [] Ufull;
  if (Uxpost) delete [] Uxpost;
  if (Uskin) delete [] Uskin;
  if (Xskin) delete [] Xskin;

}  

//-------------------------------------------------------------------------------

void split(FileNames &fn, PostOpts &po)
{

  FluidPostOp postop;
  XpostOp* xpostop = 0;

  // reading topology

  postop.initialize(fn.global, fn.geo);
  int numSkinNodes;
  int *skinNodeMap;

  // reading header from binary

  int header[3];
  if (fn.ascii) {
    xpostop = new XpostOp(fn.result);
    xpostop->isSdesign = fn.isSdesign;
    if (fn.isSdesign) {
      numSkinNodes = postop.extractSkinSurface(fn, po, skinNodeMap);
    }
    xpostop->readSolutionHeader(postop.getNumberOfNodes(), header);
  }
  else
    postop.readSolutionHeader(fn.result, header, FluidPostOp::SHARED);

  int numNodes = header[0];
  int dim = header[1];
  int numRes = header[2];

  int *activeRes;
  int numActiveRes = po.tagResults(numRes, activeRes);

  header[2] = numActiveRes;  

  // writing headers for binary

  postop.writeSolutionHeader(fn.prefix, header, FluidPostOp::DISTRIBUTED);

  // writing solution for binary

  double *Ufull = new double[numNodes * dim];

  int iActiveRes = 0;
  for (int iGlobRes=0; iGlobRes<numRes; ++iGlobRes) {

    if (activeRes[iGlobRes] == 0) continue;
    
    double tag;
    if (fn.ascii)
      xpostop->readSolution(iGlobRes, dim, tag, Ufull);
    else
      postop.readSolution(fn.result, iGlobRes, dim, tag, Ufull, FluidPostOp::SHARED);

    fprintf(stdout, " ... Processing result %d with tag %g    \r", iGlobRes+1, tag);
    fflush(stdout);

    if (fn.isSdesign)
      postop.writeSolutionSdesign(fn.prefix, numNodes, skinNodeMap, iActiveRes, dim, tag, Ufull, FluidPostOp::DISTRIBUTED);
    else
      postop.writeSolution(fn.prefix, iActiveRes, dim, tag, Ufull, FluidPostOp::DISTRIBUTED);

    ++iActiveRes;
    
  }

  fprintf(stdout, "\n");

  if (Ufull) delete [] Ufull;

}

//----------------------------------------------------------------------

void mergeStructData(FileNames &fn, int extractNum)  {

  if (extractNum > 0)
    extractNum--;

  StructPostOp postop;

  // reading header from binary
  int dataType;  // 1 for nodal, 2 for elemental
  
  postop.readClusterToSub(fn.global); // PJSA

  //int header[3];
  BinFileHandler::OffType dataOrderStart = 0;
  char *headerDescrip = postop.readSolutionHeader(fn.result, 
                        dataOrderStart, dataType);

  // read topology
  postop.initialize(/*fn.global,*/ fn.geo, dataType, dataOrderStart);

  char globFlag[MAXLINE], locFlag[MAXLINE];
  sprintf(globFlag, "wb");
  sprintf(locFlag, "rb");

  FILE *fp = new FILE;

  // write XPost Header
  postop.writeXpostHeader(fn.result, headerDescrip, fp, extractNum);

  if (dataType == 1)
    postop.processNodeSoltns(fp, extractNum);
  else 
    // postop.processNodeSoltns(fp, extractNum); 
    postop.processElemSoltns(fp, extractNum); // PJSA

}


//-------------------------------------------------------------------------------

int PostOpts::tagResults(int numRes, int *&activeRes)
{

  int numActiveRes = 0;

  activeRes = new int[numRes];

  int i;
  for (i=0; i<numRes; ++i) 
    activeRes[i] = 0;

  if (numResPicks == 0) {    
    for (i=0; i<numRes; ++i) {
      if (i % freqResPicks == 0) {
        activeRes[i] = 1;
        ++numActiveRes;
      }
    }
  }
  else {
    for (int i=0; i<numResPicks; ++i) {
      int j, k;

      if (resPicks[i][0] < 1) 
        resPicks[i][0] = 1;
      if (resPicks[i][1] > numRes) 
        resPicks[i][1] = numRes;
      for (j=resPicks[i][0], k=0; j<=resPicks[i][1]; ++j, ++k) {
        if (k % freqResPicks == 0) {
          activeRes[j-1] = 1;
          ++numActiveRes;
        }
      }
    }
  }

  fprintf(stdout, " ... Number of input result(s) = %d\n", numRes);
  fprintf(stdout, " ... Number of output result(s) = %d\n", numActiveRes);

  return numActiveRes;

}

//-------------------------------------------------------------------------------

FluidPostOp::FluidPostOp()
{
 
  numClusNodes = 0;
  numClusTets = 0;
  numClusFaces = 0;

  nodeMap = 0;

  nodes = 0;
  tets = 0;
  tetVolumeID = 0; 
  faces = 0;
  facesSurfaceID = 0;

  udist = 0;
  ushared = 0;

}

//-------------------------------------------------------------------------------

FluidPostOp::~FluidPostOp()
{

  if (numClusNodes) delete [] numClusNodes;
  if (numClusTets) delete [] numClusTets;
  if (numClusFaces) delete [] numClusFaces;

  if (nodeMap) {
    for (int i=0; i<numClusters; ++i)
      if (nodeMap[i]) delete [] nodeMap[i];
    delete [] nodeMap;
  }

  if (nodes) delete [] nodes;
  if (tets) delete [] tets;
  if (tetVolumeID) delete [] tetVolumeID;
  if (faces) delete [] faces;
  if (facesSurfaceID) delete [] facesSurfaceID;

  if (ushared) delete ushared;

  if (udist) {
    for (int i=0; i<numClusters; ++i)
      if (udist[i]) delete udist[i];
    delete [] udist;
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::initialize(const char *ncon, const char *ngeo, int dummy)
{

  nameGeo = ngeo;

  BinFileHandler fcon(ncon, "rb");

  Connectivity clusterToSub(fcon);

  numClusters = clusterToSub.csize();

  numClusNodes = new int [numClusters];
  numClusTets = new int [numClusters];
  numClusFaces = new int [numClusters];

  numNodes = 0;
  numTets = 0;
  numFaces = 0;
 
  for (int iClus=0; iClus<numClusters; ++iClus) {

    char name[MAXLINE];
    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", nameGeo, suffix);
    BinFileHandler fgeo(name, "rb");

    BinFileHandler::OffType tocOffset[3];
    fgeo.read(tocOffset, 3);

    fgeo.seek(tocOffset[0]);
    fgeo.read(numClusNodes+iClus, 1);

    fgeo.seek(tocOffset[1]);
    fgeo.read(numClusTets+iClus, 1);

    fgeo.seek(tocOffset[2]);
    fgeo.read(numClusFaces+iClus, 1);

    numNodes += numClusNodes[iClus];
    numTets += numClusTets[iClus];
    numFaces += numClusFaces[iClus];
  
  }

}

//-------------------------------------------------------------------------------

int FluidPostOp::getNumberOfNodes()
{

  if (nodeMap) 
    return numNodes;

  nodeMap = new int *[numClusters];

  int iClus;
  for (iClus=0; iClus<numClusters; ++iClus)
    nodeMap[iClus] = new int[ numClusNodes[iClus] ];

  int numOverlapNodes = numNodes;

  nodes = new double[numOverlapNodes][3];

  numNodes = 0;

  for (iClus=0; iClus<numClusters; ++iClus) {

    fprintf(stdout, " ... Reading nodes from cluster %d\n", iClus+1); 

    char name[MAXLINE];
    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", nameGeo, suffix);
    BinFileHandler fgeo(name, "rb");

    BinFileHandler::OffType start, tocOffset[3];
    fgeo.read(tocOffset, 3);

    fgeo.seek(tocOffset[0] + sizeof(int));
    fgeo.read(&start, 1);
    fgeo.seek(start);
    
    for (int i=0; i<numClusNodes[iClus]; ++i) {
      fgeo.read(&nodeMap[iClus][i], 1);  
      fgeo.read(&nodes[ nodeMap[iClus][i] ][0], 3);
      if (nodeMap[iClus][i] > numNodes) 
        numNodes = nodeMap[iClus][i];
    }

  }

  ++numNodes;

  fprintf(stdout, " ... Number of nodes (including overlap) = %d\n", numOverlapNodes);
  fprintf(stdout, " ... Number of nodes = %d\n", numNodes);

  return numNodes;

}

//-------------------------------------------------------------------------------

int FluidPostOp::getNumberOfElements()
{

  fprintf(stdout, " ... Number of elements = %d\n", numTets);

  return numTets;

}

//-------------------------------------------------------------------------------

int FluidPostOp::getNumberOfFaces()
{

  fprintf(stdout, " ... Number of boundary faces = %d\n", numFaces);

  return numFaces;

}

//-------------------------------------------------------------------------------

double (*FluidPostOp::getNodes())[3]
{

  if (!nodeMap) 
    getNumberOfNodes();

  return nodes;

}

//-------------------------------------------------------------------------------

int (*FluidPostOp::getElements())[4]
{

  if (!nodeMap)
    getNumberOfNodes();

  tets = new int[numTets][4];
  tetVolumeID = new int[numTets];

  for (int iClus=0; iClus<numClusters; ++iClus) {

    fprintf(stdout, " ... Reading elements from cluster %d\n", iClus+1); 

    char name[MAXLINE];
    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", nameGeo, suffix);
    BinFileHandler fgeo(name, "rb");

    BinFileHandler::OffType start, tocOffset[3];
    fgeo.read(tocOffset, 3);

    fgeo.seek(tocOffset[1] + sizeof(int));
    fgeo.read(&start, 1);
    fgeo.seek(start);

    for (int i=0; i<numClusTets[iClus]; ++i) {
      int type, tetMap;
      fgeo.read(&type, 1);
      fgeo.read(&tetMap, 1);   
      fgeo.read(&tetVolumeID[tetMap], 1); // CBM: porous
      fgeo.read(&tets[tetMap][0], 4);
      for (int j=0; j<4; ++j)
        tets[tetMap][j] = nodeMap[iClus][ tets[tetMap][j] ];
    }

  }

  return tets;

}

//-------------------------------------------------------------------------------

int (*FluidPostOp::getFaces())[4]
{

  if (!nodeMap)
    getNumberOfNodes();

  faces = new int[numFaces][4];
  facesSurfaceID = new int[numFaces];

  for (int iClus=0; iClus<numClusters; ++iClus) {

    fprintf(stdout, " ... Reading faces from cluster %d\n", iClus+1); 

    char name[MAXLINE];
    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", nameGeo, suffix);
    BinFileHandler fgeo(name, "rb");

    BinFileHandler::OffType start, tocOffset[3];
    fgeo.read(tocOffset, 3);

    fgeo.seek(tocOffset[2] + sizeof(int));
    fgeo.read(&start, 1);
    fgeo.seek(start);

    for (int i=0; i<numClusFaces[iClus]; ++i) {
      int type, faceMap, surface_id;
      fgeo.read(&type, 1);
      fgeo.read(&faceMap, 1);  
      fgeo.read(&faces[faceMap][0], 4);
      fgeo.read(&facesSurfaceID[faceMap], 1); // PJSA
      for (int j=0; j<3; ++j) {
        faces[faceMap][j] = nodeMap[iClus][ faces[faceMap][j] ];
      }
    }
  }

  return faces;

}

//-------------------------------------------------------------------------------

void FluidPostOp::releaseNodes()
{

  if (nodes) {
    delete [] nodes;
    nodes = 0;
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::releaseElements()
{

  if (tets) {
    delete [] tets;
    tets = 0;
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::releaseFaces()
{

  if (faces) {
    delete [] faces;
    faces = 0;
  }

}

//-------------------------------------------------------------------------------

int FluidPostOp::getNumberOfSolutions(const char *prefix)
{

  int header[3];
  readSolutionHeader(prefix, header, DISTRIBUTED);

  return header[2];

}

//-------------------------------------------------------------------------------

int FluidPostOp::getNumberOfNodalSolutions(const char *prefix)
{

  int header[3];
  readSolutionHeader(prefix, header, DISTRIBUTED);

  return header[1];

}

//-------------------------------------------------------------------------------

bool FluidPostOp::getSolution(const char *prefix, int num, TypeSol type, double &tag, double *u)
{

  int header[3];
  readSolutionHeader(prefix, header, DISTRIBUTED);

  int dim = header[1];

  if (num >= header[2]) 
    return false;

  if (!ushared)
    ushared = new Vec<double>(numNodes*dim);
  else if (ushared->size() < numNodes*dim)
    ushared->resize(numNodes*dim);

  readSolution(prefix, num, dim, tag, ushared->data(), FluidPostOp::DISTRIBUTED);

  processSolution(numNodes, dim, tag, -1.0, ushared->data(), nodes, type, u);

  return true;

}

//-------------------------------------------------------------------------------

void FluidPostOp::readSolutionHeader(const char *prefix, int *header, TypeFile type)
{

  char name[MAXLINE];

  switch (type) {
  case SHARED: {
    sprintf(name, "%s", prefix);
    break;
  }
  case DISTRIBUTED: {
    char *suffix = computeClusterSuffix(1, numClusters);
    sprintf(name, "%s%s", prefix, suffix);
    break;
  }
  }
  
  BinFileHandler fp(name, "rb");

  fp.read(header, 3);

}

//-------------------------------------------------------------------------------

void FluidPostOp::writeSolutionHeader(const char *prefix, int *header, TypeFile type)
{

  char name[MAXLINE];

  switch (type) {
  case SHARED: {
    sprintf(name, "%s", prefix);
    BinFileHandler fp(name, "w");
    fp.write(header, 3);
    break;
  }
  case DISTRIBUTED: {
    for (int iClus=0; iClus<numClusters; ++iClus) {
      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      sprintf(name, "%s%s", prefix, suffix);
      BinFileHandler fp(name, "w");
      int clusheader[3] = {numClusNodes[iClus], header[1], header[2]};
      fp.write(clusheader, 3);
    }
    break;
  }
  }

}

//-------------------------------------------------------------------------------
//    postop.readSolution(fn.result, iGlobRes, dim, tag, Ufull, FluidPostOp::DISTRIBUTED);

void FluidPostOp::readSolution(const char *prefix, int num, int dim, double &tag, 
                               double *u, TypeFile type)
{

  if (!nodes)
    getNodes();

  char name[MAXLINE];

  BinFileHandler::OffType pos;

  switch (type) {

  case SHARED: {

    sprintf(name, "%s", prefix);
    BinFileHandler fp(name, "rb");

    pos = 3*sizeof(int) + num*(sizeof(double) + numNodes*dim*sizeof(double));
    fp.seek(pos);
    fp.read(&tag, 1);
    fp.read(u, numNodes*dim);

    break;
  }

  case DISTRIBUTED: {
    
    if (!udist) {
      udist = new Vec<double>*[numClusters];
      for (int iClus=0; iClus<numClusters; ++iClus)
        udist[iClus] = new Vec<double>(numClusNodes[iClus]*dim);
    }

    for (int iClus=0; iClus<numClusters; ++iClus) {
      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      sprintf(name, "%s%s", prefix, suffix);
      BinFileHandler fp(name, "rb");

      if (udist[iClus]->size() < numClusNodes[iClus]*dim)
        udist[iClus]->resize(numClusNodes[iClus]*dim);

      pos = 3*sizeof(int) + num*(sizeof(double) + numClusNodes[iClus]*dim*sizeof(double));
      fp.seek(pos);
      fp.read(&tag, 1);
      fp.read(udist[iClus]->data(), numClusNodes[iClus]*dim);

      for (int i=0; i<numClusNodes[iClus]; ++i)
        for (int j=0; j<dim; ++j)
          u[nodeMap[iClus][i]*dim + j] = (*udist[iClus])[i*dim + j];
    }

    break;
  }
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::writeSolution(const char *prefix, int num, int dim, double tag, 
                                double *u, TypeFile type)
{

  if (!nodes)
    getNodes();

  char name[MAXLINE];

  BinFileHandler::OffType pos;

  switch (type) {

  case SHARED: {

    sprintf(name, "%s", prefix);
    BinFileHandler fp(name, "w+");

    pos = 3*sizeof(int) + num*(sizeof(double) + numNodes*dim*sizeof(double));
    fp.seek(pos);
    fp.write(&tag, 1);
    fp.write(u, numNodes*dim);

    break;
  }

  case DISTRIBUTED: {
    
    if (!udist) {
      udist = new Vec<double>*[numClusters];
      for (int iClus=0; iClus<numClusters; ++iClus)
        udist[iClus] = new Vec<double>(numClusNodes[iClus]*dim);
    }

    for (int iClus=0; iClus<numClusters; ++iClus) {
      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      sprintf(name, "%s%s", prefix, suffix);
      BinFileHandler fp(name, "w+");

      if (udist[iClus]->size() < numClusNodes[iClus]*dim)
        udist[iClus]->resize(numClusNodes[iClus]*dim);

      for (int i=0; i<numClusNodes[iClus]; ++i)
        for (int j=0; j<dim; ++j)
          (*udist[iClus])[i*dim + j] = u[nodeMap[iClus][i]*dim + j];

      pos = 3*sizeof(int) + num*(sizeof(double) + numClusNodes[iClus]*dim*sizeof(double));
      fp.seek(pos);
      fp.write(&tag, 1);
      fp.write(udist[iClus]->data(), numClusNodes[iClus]*dim);

    }

    break;
  }
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::writeSolutionSdesign(const char *prefix, int numSkinNodes, int* skinNodeMap, 
                                       int num, int dim, double tag, double *u, TypeFile type)
{

  if (!nodes)
    getNodes();

  char name[MAXLINE];
  Vec<double> ufull(numNodes*dim, 0.0);

/* YC: This should be commented out.
    for (int i=0; i<numNodes; ++i)
      for (int j=0; j<dim; ++j) {
        ufull[i*dim+j] = nodes[i][j];
      }
*/

  for(int i=0; i<numSkinNodes; ++i) 
    for(int j=0; j<dim; ++j) {
      ufull[skinNodeMap[i]*dim+j] += u[i*dim+j];
    }


  BinFileHandler::OffType pos;

  switch (type) {

  case SHARED: {

    fprintf(stderr,"FluidPostOp::writeSolutionSdesign for shared memory has not been implemented.\n");
    break;
  }

  case DISTRIBUTED: {
    
    if (!udist) {
      udist = new Vec<double>*[numClusters];
      for (int iClus=0; iClus<numClusters; ++iClus)
        udist[iClus] = new Vec<double>(numClusNodes[iClus]*dim);
    }

    for (int iClus=0; iClus<numClusters; ++iClus) {
      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      sprintf(name, "%s%s", prefix, suffix);
      BinFileHandler fp(name, "w+");

      if (udist[iClus]->size() < numClusNodes[iClus]*dim)
        udist[iClus]->resize(numClusNodes[iClus]*dim,0.0);

      for (int i=0; i<numClusNodes[iClus]; ++i)
        for (int j=0; j<dim; ++j) {
          (*udist[iClus])[i*dim + j] = ufull[(nodeMap[iClus][i])*dim + j];
        }

      pos = 3*sizeof(int) + num*(sizeof(double) + numClusNodes[iClus]*dim*sizeof(double));
      fp.seek(pos);
      fp.write(&tag, 1);
      fp.write(udist[iClus]->data(), numClusNodes[iClus]*dim);

    }

    break;
  }
  }

}

//-------------------------------------------------------------------------------

void FluidPostOp::writeSolutionEnsight(const char *prefix, int num, int dim, double tag, 
                                       double *u, TypeFile type)
{

  if (!nodes)
    getNodes();

  char name[256];

  switch (type) {

  case SHARED: {
    float *u_j = new float[numNodes];

    if (dim==1 || dim==3) {

      // open file
      sprintf(name, "%s_%.3d.var", prefix, num);    
      fprintf(stdout, "... Writing Ensight variable file \'%s\'\n", name);
      EnsightFileHandler outputVar(name, "wb");
      
      // write header
      char tag_line[MAXLINE];
      sprintf(tag_line, "Tag %e (%d)", tag, num);
      outputVar.header(tag_line);
      
      // start new part
      outputVar.part("coordinates");
      
      for (int j=0; j<dim; ++j) {
        for (int i=0; i<numNodes; ++i)
          u_j[i] = u[i*dim + j];
  
        outputVar.write(u_j, numNodes, 1);
      }

    } else {

      for (int j=0; j<dim; j++) {

        // open file
        sprintf(name, "%s_v%d_%.3d.var", prefix, j+1, num);    
        fprintf(stdout, "... Writing Ensight variable file \'%s\'\n", name);
        EnsightFileHandler outputVar(name, "wb");
    
        // write header
        char tag_line[MAXLINE];
        sprintf(tag_line, "Tag %e (%d)", tag, num);
        outputVar.header(tag_line);
    
        // start new part
        outputVar.part("coordinates");
    
        for (int i=0; i<numNodes; ++i)
          u_j[i] = u[i*dim + j];
    
        outputVar.write(u_j, numNodes, 1);
 
      }
    }

    delete [] u_j;

    break;
  }

  case DISTRIBUTED: {
    
    // open file
    sprintf(name, "%s.var", prefix);    
    fprintf(stdout, "... Writing Ensight variable file \'%s\'\n", name);
    EnsightFileHandler outputVar(name, "wb");
    
    // write header
    char tag_line[MAXLINE];
    sprintf(tag_line, "Tag %e", tag);
    outputVar.header(tag_line);

    for (int iClus=0; iClus<numClusters; ++iClus) {
      // start new part
      outputVar.part("coordinates");
      
      float *u_j = new float[numClusNodes[iClus]];
      for (int j=0; j<dim; ++j) {
        for (int i=0; i<numClusNodes[iClus]; ++i)
          u_j[i] = u[nodeMap[iClus][i]*dim + j];
  
        outputVar.write(u_j, numClusNodes[iClus]);
      }
      delete [] u_j;
    }

    break;
  }
  }

}

//-------------------------------------------------------------------------------

int FluidPostOp::extractSkinSurface(FileNames &fn, PostOpts &po, int *&skinNodeMap)
{

  if (!nodes)
    getNodes();

  if (!faces)
    getFaces();

  int numSkinNodes = 0;
  skinNodeMap = new int[numNodes];

  int numSkinFaces = 0;
  Vec<int> skinFaceMap(numFaces);
  Vec<int> nodeTag(numNodes); 

  nodeTag = -1;

  char nodemap_name[MAXLINE];
  sprintf(nodemap_name, "%s.nodemap", fn.nodemap);
  FILE *nodemapFile = fopen(nodemap_name, "w");
  fprintf(nodemapFile, "node map ascii\n");

  int k, i;
  for (k=0; k<po.numBCs; ++k) {
    for (i=0; i<numFaces; ++i) {
      if (faces[i][3] == po.bcs[k]) {
        for (int j=0; j<3; ++j) {
          if (nodeTag[ faces[i][j] ] == -1) {
            skinNodeMap[numSkinNodes] = faces[i][j];
            nodeTag[ faces[i][j] ] = numSkinNodes;
//            fprintf(stderr,"skinNodeMap[%d] = %d\n", numSkinNodes, skinNodeMap[numSkinNodes]);
            ++numSkinNodes;
          }
        }
        skinFaceMap[numSkinFaces] = i;
        ++numSkinFaces;
      }
    }
  }

  for (k=0; k<numSkinNodes; ++k) {
    fprintf(nodemapFile, "%d %d\n", k+1, skinNodeMap[k]);
  }


  fprintf(stdout, " ... Number of skin nodes = %d\n", numSkinNodes);
  fprintf(stdout, " ... Number of skin faces = %d\n", numSkinFaces);

  char name[MAXLINE];
  sprintf(name, "%s.xpost", fn.skin);
  FILE *skinFile = fopen(name, "w");

#ifdef VRML
  fprintf(skinFile,"#VRML V1.0 ascii\n");
  fprintf(skinFile,"\n");
  fprintf(skinFile,"\n");
  fprintf(skinFile,"Separator {\n");
  fprintf(skinFile," Coordinate3 {\n");
  fprintf(skinFile," point [\n");
#endif
#ifdef STL
  fprintf(skinFile,"solid ICEM CFD/CAE part\n");
#endif
#ifdef XPOST
  fprintf(skinFile,"Nodes %s\n", fn.nodes);
#endif

  SVec<double,3> Xskin(numSkinNodes);

  for (i=0; i<numSkinNodes; ++i) {
    for (int j=0; j<3; ++j)
      Xskin[i][j] = nodes[ skinNodeMap[i] ][j];
#ifdef VRML
    fprintf(skinFile,"%g %g %g,\n", Xskin[i][0], Xskin[i][1], Xskin[i][2]);
#endif
#ifdef XPOST
    fprintf(skinFile,"%d %g %g %g\n", i+1, Xskin[i][0], Xskin[i][1], Xskin[i][2]);
#endif
  }
 
#ifdef SYMMETRY
  for (i=0; i<numSkinNodes; ++i)
    fprintf(skinFile,"%d %g %g %g\n", i+numSkinNodes+1, -Xskin[i][0], Xskin[i][1], Xskin[i][2]);
#endif

#ifdef VRML
  fprintf(skinFile, " ]\n");
  fprintf(skinFile, " }\n");
  fprintf(skinFile, " IndexedFaceSet {\n");
  fprintf(skinFile, " coordIndex [\n");
#endif
#ifdef XPOST
  fprintf(skinFile, "Elements %s using %s\n", fn.elements, fn.nodes);
#endif

  for (i=0; i<numSkinFaces; ++i) {
#ifdef VRML
    fprintf(skinFile,"%d, %d, %d, -1,\n",
            nodeTag[ faces[ skinFaceMap[i] ][0] ],
            nodeTag[ faces[ skinFaceMap[i] ][1] ],
            nodeTag[ faces[ skinFaceMap[i] ][2] ]);
#endif
#ifdef STL
    int num[3] = {nodeTag[ faces[ skinFaceMap[i] ][0] ], 
                  nodeTag[ faces[ skinFaceMap[i] ][1] ],
                  nodeTag[ faces[ skinFaceMap[i] ][2] ]};

    Vec3D x[3] = {Xskin[ num[0] ], Xskin[ num[1] ], Xskin[ num[2] ]};
    Vec3D n = 0.5 * ((x[2] - x[0]) ^ (x[1] - x[0]));
    n *= - 1.0 / sqrt(n*n);

    fprintf(skinFile,"  facet normal %e %e %e\n", n[0], n[1], n[2]);
    fprintf(skinFile,"    outer loop\n");
    fprintf(skinFile,"      vertex %e %e %e\n", x[0][0], x[0][1], x[0][2]);
    fprintf(skinFile,"      vertex %e %e %e\n", x[1][0], x[1][1], x[1][2]);
    fprintf(skinFile,"      vertex %e %e %e\n", x[2][0], x[2][1], x[2][2]);
    fprintf(skinFile,"    endloop\n");
    fprintf(skinFile,"  endfacet\n");
#endif
#ifdef XPOST
    fprintf(skinFile,"%d 4 %d %d %d\n", i + 1, 
            nodeTag[ faces[ skinFaceMap[i] ][0] ] + 1,
            nodeTag[ faces[ skinFaceMap[i] ][1] ] + 1,
            nodeTag[ faces[ skinFaceMap[i] ][2] ] + 1);
#endif
  }

#ifdef SYMMETRY
  for (i=0; i<numSkinFaces; ++i) {
    fprintf(skinFile,"%d 4 %d %d %d\n", i + numSkinFaces + 1, 
            nodeTag[ faces[ skinFaceMap[i] ][0] ] + numSkinNodes + 1,
            nodeTag[ faces[ skinFaceMap[i] ][2] ] + numSkinNodes + 1,
            nodeTag[ faces[ skinFaceMap[i] ][1] ] + numSkinNodes + 1);
  }
#endif
  
#ifdef VRML
  fprintf(skinFile," ]\n");
  fprintf(skinFile," }\n");
  fprintf(skinFile,"}\n");
#endif
#ifdef STL
  fprintf(skinFile,"endsolid\n");
#endif

  fclose(skinFile);
  fclose(nodemapFile);
  return numSkinNodes;

}

//-------------------------------------------------------------------------------

void FluidPostOp::writeTopology(FileNames& fn)
{

  if (!nodes)
    getNodes();
  
  if (!faces)
    getFaces();

  if (!tets)
    getElements();

  FILE* fp = fopen("topo.xpost", "w");

  fprintf(fp, "Nodes %s\n", fn.nodes);
  int i;
  for (i=0; i<numNodes; ++i)
    fprintf(fp, "%d %e %e %e\n", i+1, nodes[i][0], nodes[i][1], nodes[i][2]);

  fprintf(fp, "Elements %s using %s\n", fn.elements, fn.nodes);
  for (i=0; i<numTets; ++i)
    if(tetVolumeID[i] == -1)
    fprintf(fp, "%d 5 %d %d %d %d\n", i+1, tets[i][0]+1, tets[i][1]+1, tets[i][2]+1, tets[i][3]+1);

  int MAX_VOLUME_ID = 0;
  for (i=0; i<numTets; ++i)
    if (tetVolumeID[i] > MAX_VOLUME_ID) MAX_VOLUME_ID = tetVolumeID[i];


  for(int j=0; j<=MAX_VOLUME_ID; ++j) {
    bool first_time = true;
    for (i=0; i<numTets; ++i) {
      if (tetVolumeID[i] == j) {
        if(first_time) {
          fprintf(fp, "Elements PorousVolume_%d using FluidNodes\n", tetVolumeID[i]);
          first_time = false;
        }
        fprintf(fp, "%d 5 %d %d %d %d\n",
                i+1, tets[i][0]+1, tets[i][1]+1, tets[i][2]+1, tets[i][3]+1);
      }
    }
  }

#define MIN_TYPE_BC -5
#define MAX_TYPE_BC 10

  const char** type_name = new const char*[MAX_TYPE_BC - MIN_TYPE_BC + 1];
  type_name -= MIN_TYPE_BC;
  for (i=MIN_TYPE_BC; i<=MAX_TYPE_BC; ++i)
    type_name[i] = "InactiveSurface";

  type_name[-5] = "OutletMoving";
  type_name[-4] = "InletMoving";
  type_name[-3] = "StickMoving";
  type_name[-2] = "SlipMoving";
  type_name[0] = "Internal";
  type_name[2] = "SlipFixed";
  type_name[3] = "StickFixed";
  type_name[4] = "InletFixed";
  type_name[5] = "OutletFixed";
  type_name[6] = "Symmetry"; // PJSA
  type_name[10] = "Missing";

//  for (int ii=MIN_TYPE_BC; ii<=MAX_TYPE_BC; ++ii) {
//    int count = 0;
//    for (i=0; i<numFaces; ++i)
//      if (faces[i][3] == ii)
//        ++count;
//    if (count) {
//      fprintf(fp, "Elements %sSurface using %s\n", type_name[ii], fn.nodes);
//      count = 0;
//      for (i=0; i<numFaces; ++i)
//        if (faces[i][3] == ii)
//          fprintf(fp, "%d 4 %d %d %d\n", ++count, faces[i][0]+1, faces[i][1]+1, faces[i][2]+1);
//    }
//  }


// CBM
  int MAX_SURFACE_ID = 0;
  for (i=0; i<numFaces; ++i)
    if(facesSurfaceID[i] > MAX_SURFACE_ID) MAX_SURFACE_ID = facesSurfaceID[i];

  for (int ii=MIN_TYPE_BC; ii<=MAX_TYPE_BC; ++ii) {
    for (int k=0; k<=MAX_SURFACE_ID; ++k) {
      int nelements = 0;
      bool first_time = true;
      for (i=0; i<numFaces; ++i) {
        if (faces[i][3] == ii && facesSurfaceID[i] == k) {
          if(first_time) {
            fprintf(fp, "Elements %sSurface_%d using FluidNodes\n", type_name[ii], facesSurfaceID[i]);
            first_time = false;
          }
          fprintf(fp, "%d 4 %d %d %d\n",
                  ++nelements, faces[i][0]+1, faces[i][1]+1, faces[i][2]+1);
        }
      }
    }
  }


  fclose(fp);

}

//-------------------------------------------------------------------------------

StructPostOp::StructPostOp()
{
  numRes = 0;
  dim = 0;

  numClusItems = 0;

  numItems = 0;

  clusToGlMap = 0;

  elemTypeNumNodesMap = 0;
  glElemLengths = 0;

  clusLengths = 0; // PJSA
}


void StructPostOp::readClusterToSub(const char *globFile)
{
  // get connectivity info.
  BinFileHandler conFile(globFile, "rb");
  Connectivity clusterToSub(conFile);
  numClusters = clusterToSub.csize();
}

//--------------------------------------------------------------------

void StructPostOp::initialize(/*char *globFile,*/ const char *geometryFile, 
                              int dataType, BinFileHandler::OffType &orderStart)  {

  dataType--;  // sets 0 for nodal, 1 for elemental

  BinFileHandler::OffType start, tocOffset[3];

  char name[MAXLINE]; 

/* PJSA: moved to readClusterToSub, need numClusters in readSolutionHeader
  // get connectivity info.
  BinFileHandler conFile(globFile, "rb");
  Connectivity clusterToSub(conFile);
  numClusters = clusterToSub.csize();
*/

  // getting the nonoverlapping sizes
  numClusItems = new int [numClusters];
  clusToGlMap = new int *[numClusters];
  clusLengths = new int [numClusters]; // PJSA

  // get number of things per cluster
  int iClus;
  for (iClus = 0; iClus < numClusters; ++iClus) {

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", geometryFile, suffix);
    BinFileHandler geoFile(name, "rb");

    geoFile.read(tocOffset, 3);
    geoFile.seek(tocOffset[dataType]);
    geoFile.read(numClusItems+iClus, 1);

    clusToGlMap[iClus] = new int[ numClusItems[iClus] ];

    clusLengths[iClus] = 0; // PJSA
  }

  int numOverlapNodes = 0;

  for (iClus = 0; iClus < numClusters; ++iClus)
    numOverlapNodes += numClusItems[iClus];

  if (dataType == 1)  {  // PJSA
    int totElems = 0;
    for (iClus = 0; iClus < numClusters; ++iClus)
      totElems += numClusItems[iClus];

    glElemLengths = new int [totElems+1];
  }

  // create the cluster to global item mapping
  for (iClus = 0; iClus < numClusters; ++iClus) {

    fprintf(stdout, " ... Reading geometry cluster %d\n", iClus+1);

    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    sprintf(name, "%s%s", geometryFile, suffix);
    BinFileHandler geoFile(name, "rb");

    geoFile.read(tocOffset, 3);

    geoFile.seek(tocOffset[dataType] + sizeof(int));
    geoFile.read(&start, 1);
    geoFile.seek(start);

    if (dataType == 0)  {
      readNodes(geoFile, iClus);
      fprintf(stdout, " ... Number of nodes (including overlap) = %d\n", 
              numOverlapNodes);
      fprintf(stdout, " ... Number of nodes = %d\n", numItems);
    }
    else if (dataType == 1)  {
      readElems(geoFile, iClus, glElemLengths);
      fprintf(stderr," ... Number of elements = %d\n", numItems); // PJSA
    }
    else  {
      fprintf(stderr,"*** ERROR: Data Type Error %d\n", dataType);
      exit (1);
    }
  }

  // get node or elem order from result files
  // locToClusMap = new int *[numClusters];
  locToGlMap = new int *[numClusters]; // PJSA
  nClusData = new int[numClusters];
  soltnStart = new BinFileHandler::OffType[numClusters];

  for (iClus = 0; iClus < numClusters; iClus++)  {
    char *suffix = computeClusterSuffix(iClus+1, numClusters);
    char name[MAXLINE];
    sprintf(name, "%s%s", nameRes, suffix);
    BinFileHandler resFile(name, "rb");

    // get number of master cluster nodes
    resFile.seek(orderStart);
    resFile.read(nClusData+iClus, 1);
    //locToClusMap[iClus] = new int[nClusData[iClus]];
    locToGlMap[iClus] = new int[nClusData[iClus]];

    // get cluster data order
    resFile.seek(orderStart+3*sizeof(int));
    //resFile.read(locToClusMap[iClus], nClusData[iClus]);
    resFile.read(locToGlMap[iClus], nClusData[iClus]);
    soltnStart[iClus] = resFile.tell();
  }

}

//--------------------------------------------------------------------

void StructPostOp::readNodes(BinFileHandler &geoFile, int iClus)  {

  bool found = false;
  for (int cItem = 0; cItem < numClusItems[iClus]; ++cItem) {

    geoFile.read(&clusToGlMap[iClus][cItem], 1);

    double coords[3];
    geoFile.read(coords, 3);
    if (clusToGlMap[iClus][cItem] >= numItems) {
      numItems = clusToGlMap[iClus][cItem];
      found = true;
    }

  }
  if(found) ++numItems;

}

//--------------------------------------------------------------------

void StructPostOp::readElems(BinFileHandler &geoFile, int iClus, 
                             int *glElemLengths)  {

  bool found = false;
  // get mapping for elem type to numNodes
  setElemTypeMap();

  // read in elem type, global elem number, and its node numbers
  int elemNodes[10];
  for (int cItem = 0; cItem < numClusItems[iClus]; ++cItem) {

    int elemType;
    geoFile.read(&elemType, 1);

    geoFile.read(&clusToGlMap[iClus][cItem], 1);

    int numNodes = elemTypeNumNodesMap[elemType];

    if (numNodes < 0)  {
      fprintf(stderr,"*** ERROR: Elem type %d for %d not supported\n",
                      elemType, clusToGlMap[iClus][cItem]);
      exit (-1);
    }
    glElemLengths[clusToGlMap[iClus][cItem]] = numNodes;

    clusLengths[iClus] += numNodes;  // PJSA

    if (clusToGlMap[iClus][cItem] >= numItems) {
      numItems = clusToGlMap[iClus][cItem];
      found = true;
    }

    int volumeId;
    geoFile.read(&volumeId, 1);

    // skip to next element
    geoFile.read(elemNodes, numNodes);

  }
  if(found) ++numItems;
}

//---------------------------------------------------------------------

char *StructPostOp::readSolutionHeader(const char *prefix, 
                                       BinFileHandler::OffType &start, int &dataType)  {

  nameRes = prefix;
  char name[MAXLINE];

  char *suffix = computeClusterSuffix(1, numClusters);
  sprintf(name, "%s%s", prefix, suffix);

  BinFileHandler fp(name, "rb");

  // read dataType
  int type;
  fp.read(&type, 1);

  if (type != 1 && type != 2)  {
    fprintf(stderr,"Bad data type: %d \n", dataType);
    exit(1);
  }
  dataType = type;

  // read header description size
  fp.read(&headerSize, 1);
  
  // allocate head description array
  char *headerDescrip = new char[headerSize+1];

  // read header description
  fp.read(headerDescrip, headerSize);
  headerDescrip[headerSize] = '\0';

  start = fp.tell();

  // read result file (num. data, dimension, num. results)
  int header[3];
  fp.read(header, 3);
  dim = header[1];
  numRes = header[2];

  fprintf(stderr,"Data file has %d items w/ dim %d w/ %d results starting here: %ld\n", header[0], dim, numRes, fp.tell());
  return headerDescrip;
}

//---------------------------------------------------------------------

void StructPostOp::writeXpostHeader(const char *prefix, const char *descrip, 
                                    FILE *&fp, int extractNum)  {

  if (prefix[0] == 0) return;

  char name[MAXLINE];
  sprintf(name, "%s.xpost", prefix);

  fp = fopen(name, "w");

  fprintf(fp, descrip);  // descrip already contains numItems

  if (extractNum >= 0)
    fprintf(fp, " %d\n", extractNum);
/*
  else
    // print total number of nodes
    fprintf(fp, "%d\n", numItems);  
*/
}

//------------------------------------------------------------

void StructPostOp::setElemTypeMap()  {

  // build elem type to num nodes map
  elemTypeNumNodesMap = new int[112];
  for (int i = 0; i < 112; i++)
    elemTypeNumNodesMap[i] = -1;

  elemTypeNumNodesMap[ 1] =  2;
  elemTypeNumNodesMap[ 2] =  4;
  elemTypeNumNodesMap[ 3] =  4;
  elemTypeNumNodesMap[ 4] =  4;
  elemTypeNumNodesMap[ 6] =  2;
  elemTypeNumNodesMap[ 7] =  3;
  elemTypeNumNodesMap[ 8] =  3;
  elemTypeNumNodesMap[ 9] =  2;
  elemTypeNumNodesMap[10] =  4;
  elemTypeNumNodesMap[11] =  1;
  elemTypeNumNodesMap[17] =  8;
  elemTypeNumNodesMap[18] =  4;
  elemTypeNumNodesMap[19] =  3;
  elemTypeNumNodesMap[20] =  3;
  elemTypeNumNodesMap[21] =  2;
  elemTypeNumNodesMap[22] =  2;
  elemTypeNumNodesMap[23] =  4;
  elemTypeNumNodesMap[24] =  5;
  elemTypeNumNodesMap[26] = 10;
  elemTypeNumNodesMap[30] =  4;
  elemTypeNumNodesMap[31] =  4;
  elemTypeNumNodesMap[35] =  3;
  elemTypeNumNodesMap[36] =  3;
  elemTypeNumNodesMap[40] =  4;
  elemTypeNumNodesMap[41] =  4;
  elemTypeNumNodesMap[52] =  8;
  elemTypeNumNodesMap[60] =  4;
  elemTypeNumNodesMap[61] =  3;
  elemTypeNumNodesMap[88] =  4; //HB
  elemTypeNumNodesMap[111] = 2;
}

//----------------------------------------------------------------------

void StructPostOp::processNodeSoltns(FILE *&fp, int extractNum)  {
  
  // get the solution vectors and combine
  double *globalData = new double[numItems * dim];

  for (int iRes = 0; iRes < numRes; iRes++)  {

    double tag;

    int iClus;
    for (iClus = 0; iClus < numClusters; iClus++)  {

      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      char name[MAXLINE];
      sprintf(name, "%s%s", nameRes, suffix);
      BinFileHandler resFile(name, "rb");

      // seek to start of solution
      resFile.seek(soltnStart[iClus] + iRes*(nClusData[iClus]*dim+1)*sizeof(double));

      // read time tag
      resFile.read(&tag, 1);

      // read in solution vector
      double *clusData = new double[dim * nClusData[iClus]];
      BinFileHandler::OffType startPos = resFile.tell();
      resFile.read(clusData, dim*nClusData[iClus]);

      // assign cluster data to global data array
      for (int iData = 0; iData < nClusData[iClus]; iData++)  {

        // get global node
        // int glNode = clusToGlMap[iClus][ locToClusMap[iClus][iData] ];
        int glNode = locToGlMap[iClus][iData];
        for (int iComp = 0; iComp < dim; iComp++)  {
          // fprintf(stderr,"Assigning glNode[%d][%d] = %e, glIndex = %d\n", glNode, iComp, 
          //         clusData[iData*dim + iComp], glNode*dim + iComp);
          globalData[glNode*dim + iComp] = clusData[iData*dim + iComp];
        }

      }
    }

    // write solution to text XPost Format
    writeXpostNodeSolution(tag, globalData, fp, extractNum);

  }
}

//-----------------------------------------------------------------------

void StructPostOp::processElemSoltns(FILE *&fp, int extractNum)  {

  int iClus;

  // get totalsize of global data and create offset map
  int globLength = 0;
  glElemOffsets = new int[numItems];
  for(int iElem = 0; iElem < numItems; ++iElem) {
    glElemOffsets[iElem] = globLength;
    globLength += glElemLengths[iElem];
  }
  double *globalData = new double[globLength];
    
  for (int iRes = 0; iRes < numRes; iRes++)  {

    double tag;

    for (iClus = 0; iClus < numClusters; iClus++)  {

      char *suffix = computeClusterSuffix(iClus+1, numClusters);
      char name[MAXLINE];
      sprintf(name, "%s%s", nameRes, suffix);
      BinFileHandler resFile(name, "rb");

      // seek to start of solution
      resFile.seek(soltnStart[iClus] + iRes*(1+clusLengths[iClus])*sizeof(double));

      // read time tag
      resFile.read(&tag, 1);

      // read in solution vector
      double *clusData = new double[clusLengths[iClus]];
      BinFileHandler::OffType startPos = resFile.tell();
      resFile.read(clusData, clusLengths[iClus]);

      // assign cluster data to global data array
      int offset = 0;
      for (int iElem = 0; iElem < nClusData[iClus]; iElem++)  {

        // get global element
        // int glElem = clusToGlMap[iClus][ locToClusMap[iClus][iElem] ];
        int glElem = locToGlMap[iClus][iElem];
        for (int iNode = 0; iNode < glElemLengths[glElem]; iNode++)  {
          // fprintf(stderr,"Assigning glElem[%d][%d] = %e, glIndex = %d\n", glElem, iNode, 
          //         clusData[offset], glElemOffsets[glElem]+iNode);
          globalData[glElemOffsets[glElem] + iNode] = clusData[offset];
          offset++;
        }

      }
    }

    // write solution to text XPost Format
    writeXpostElemSolution(tag, globalData, fp, extractNum);
  }
  
  delete [] globalData;
}

//-------------------------------------------------------------------------

void StructPostOp::writeXpostNodeSolution(double tag, double *data, FILE *&fp,
                                          int extractNum)  {

  // write time tag
  fprintf(fp, "  % *.*E\n", 10, 4, tag);

  // write solution
  if(extractNum < 0)  {
    for(int i = 0; i < numItems; ++i)  {
      if(dim==6) fprintf(fp, "%d ", i+1);  // PJSA: write node for output6
      for(int iComp = 0; iComp < dim; iComp++)
        fprintf(fp, " %  *.*E", 10, 4, data[i*dim + iComp]);
      fprintf(fp, "\n");
    } 
  }
  else {
    for(int iComp = 0; iComp < dim; iComp++)
      fprintf(fp, " %  *.*E ", 10, 4, data[extractNum*dim + iComp]);
    fprintf(fp, "\n");
  }
}

//-------------------------------------------------------------------------

void StructPostOp::writeXpostElemSolution(double tag, double *data, FILE *&fp,
                   int extractNum)  {

  // write time tag
  fprintf(fp, "  % *.*E\n", 10, 4, tag);

  // write solution
  if (extractNum < 0)  {
    for (int i = 0; i < numItems; ++i)  {

      for (int iNode = 0; iNode < glElemLengths[i]; iNode++)
        fprintf(fp, " %  *.*E", 10, 4, data[glElemOffsets[i] + iNode]);

      fprintf(fp, "\n");
    }
  }
  else  {

    for (int iNode = 0; iNode < glElemLengths[extractNum]; iNode++)
      fprintf(fp, "%  *.*E ", 10, 4, data[glElemOffsets[extractNum] + iNode]);
    fprintf(fp, "\n");
  }
}

//-------------------------------------------------------------------------------

XpostOp::XpostOp(const char* name)
{

  size = 0;
  current = 0;
  fname = name;
  isSdesign = false;
}

//-------------------------------------------------------------------------------

XpostOp::~XpostOp()
{

}

//-------------------------------------------------------------------------------

void XpostOp::readSolutionHeader(int _size, int header[3])
{

  fp = fopen(fname, "r");
  if (!fp) {
    fprintf(stderr, "*** Error: file \'%s\' not found\n", fname);
    exit(1);
  }

  char line[MAXLINE];      
  fgets(line, MAXLINE, fp);
  fgets(line, MAXLINE, fp);
  sscanf(line, "%d", &size);
  header[0] = size;
  if(!isSdesign) {
    if (size != _size) {
      fprintf(stderr, "*** Error: wrong size (%d vs. %d)\n", _size, size);
      exit(1);
    }
  }
  int nlines = 0;
  while (fgets(line, MAXLINE, fp) != 0)
    ++nlines;
  header[2] = nlines/(size+1);
  fclose(fp);

  fp = fopen(fname, "r");
  int i;
  for (i=0; i<4; ++i)
    fgets(line, MAXLINE, fp);
  int linelen = strlen(line);
  i = 0;
  header[1] = 0;
  char item[MAXLINE];
  do {
    if (line[i] != ' ') {
      ++header[1];
      sscanf(line+i, "%s", item);
      i += strlen(item) - 1;
    }
    ++i;
  } while (i != (linelen - 1));
  fclose(fp);

}

//-------------------------------------------------------------------------------

void XpostOp::readSolution(int num, int dim, double& tag, double* sol)
{

  char line[MAXLINE], item[MAXLINE];

  if (current == 0) {
    fp = fopen(fname, "r");
    fgets(line, MAXLINE, fp);
    fgets(line, MAXLINE, fp);
  }
 
  int i;
  for (i=0; i<(num-current)*(size+1); ++i)
    fgets(line, MAXLINE, fp);

  fgets(line, MAXLINE, fp);
  sscanf(line, "%lf", &tag);
  for (int ii=0; ii<size; ++ii) {
    fgets(line, MAXLINE, fp);
    i = 0;
    int j = 0;
    int linelen = strlen(line);
    do {
      if (line[i] != ' ') {
        sscanf(line+i, "%s", item);
        sol[ii*dim+j] = atof(item);
        i += strlen(item) - 1;
        ++j;
      }
      ++i;
    } while (i != (linelen - 1));
  }

  current = num + 1;

}

//-------------------------------------------------------------------------------
