#include <alloca.h>
#include <cstdio>
#include <cstdlib>
#include "Domain.h"
#include "Decomp3D.h"
#include "Tetrahedron3D.h"
#include "BinFileHandler.h"

#include "FourNodeQuad.h"
#include "TwoNodeTruss.h"
#include "Therm3DQuad.h"
#include "Triangle3.h"
#include "EulerBeam.h"
#include "TimoshenkoBeam.h"
#include "ThreeNodeShell.h"
#include "FourNodeShell.h"
#include "Therm2NodeBar.h"
#include "ThermQuadGal.h"
#include "TorSpring.h"
#include "EightNodeBrick.h"
#include "ShearPanel.h"
#include "Membrane.h"

#include "Compo3NodeShell.h"
#include "TransSprlink.h"
#include "RotnSprlink.h"
#include "Tetrahedral.h"
#include "Pentahedral.h"
#include "TenNodeTetrahedral.h"
#include "HelmQuadGal.h"
#include "HelmQuadGls.h"
#include "HelmTri3Gal.h"
#include "HelmTri3Gls.h"
#include "TetraHelmGal.h"
#include "TetraHelmGLS.h"
#include "TetraTherm.h"
#include "ThermBrick.h"

#include "HelmIsoParamHexa.h"
#include "HelmIsoParamTetra.h"

#include "BlockAlloc.h"

extern int yyparse(void);

Domain *domain;

//--------------------------------------------------------------------

StructureDomain::StructureDomain(int numClus) : Domain(numClus), 
			cframes(0), layInfo(0), coefData(0)  
{
/* PJSA: moved to domain class
  // set bc data to zero
  numIDis6 = 0;
  numIDis = 0; 
  numIVel = 0; 
  numDirichlet = 0; 
  numNeuman = 0; 
  //numSommer = 0;
  //numLMPC = 0;
  numConvBC = 0; 
  numITemp = 0;
  //numComplexDirichlet = 0; 
  //numComplexNeuman = 0; 

  // set bc pointers to zero
  dbc = 0;
  nbc = 0;
  cvbc = 0;
  iDis = 0;
  iDis6 = 0;
  iVel = 0;
  iTemp = 0;
*/

  // set per cluster members to zero
  dbcPerClus = 0;
  nbcPerClus = 0;
  cvbcPerClus = 0;
  iDisPerClus = 0;
  iDis6PerClus = 0;
  iVelPerClus = 0;
  iTempPerClus = 0;

  // initialize properties
  numprops = 0;
  sProps = 0;
  na = 0;
  namax = 0;
  attrib = 0;
  numlayInfo = 0;
  numCFrames = 0;
  numCoefData = 0;
}

//--------------------------------------------------------------------

void StructureDomain::getGeometry(const char *file)  {

  // initialize number of nodes
  numNode = 0;
  numElem = 0;

  // set Attribute array
  na = 0;
  namax = 256;
  attrib  = new Attrib[namax];

  // set Structure Property Array
  numprops=32;
  sProps = new StructProp[numprops];

    firstEFData = 0;
  numMatchData = 0;

  numprops = 0;
  numCFrames = 0;

 
  // use parser to read fem input file
  domain = this;
  fprintf(stderr, "... Parsing the input file: %s \n", file);
  FILE *nin = freopen(file,"r",stdin);
  int error;
  error = yyparse();
  fprintf(stderr, "... Completed parsing input file \n");

  fprintf(stderr, "... Num Elems: %d, Num Nodes: %d\n", numElem, numNode);

  // Check if Input file had any errors
  if(error) {
    fprintf(stderr," *******************************************************\n");
    fprintf(stderr," *** ERROR: Structure Input file incompletely parsed ***\n");
    fprintf(stderr," *******************************************************\n");
    exit(error);
  }

}
//--------------------------------------------------------------------


void StructureDomain::addNode(int nd, double p[3])
{
 Point3D &n = node[nd];
 n[0] = p[0];
 n[1] = p[1];
 n[2] = p[2];
 if(nd >= numNode) numNode = nd+1;
}

//-----------------------------------------------------------------------

int StructureDomain::setDirichlet(int _numDirichlet, BCond *_dbc)
{
 // dbc = dirichlet boundary conditions
 if(dbc) {

   // Allocate memory for correct number of dbc
   BCond *nd = new BCond[numDirichlet+_numDirichlet];

   // copy old dbc
   int i;
   for(i=0; i < numDirichlet; ++i)
      nd[i] = dbc[i];

   // copy new dbc
   for(i=0; i<_numDirichlet; ++i)
     nd[i+numDirichlet] = _dbc[i];

   // set correct number of dbc
   numDirichlet += _numDirichlet;

   // delete old array of dbc
   delete [] dbc;
  
   // set new pointer to correct number of dbc
   dbc = nd;

 } else {
   numDirichlet = _numDirichlet;
   dbc          = _dbc;
 }
 return 0;
}

//-----------------------------------------------------------------------

int StructureDomain::setNeuman(int _numNeuman, BCond *_nbc)
{
 numNeuman = _numNeuman;
 nbc       = _nbc;
 return 0;
}

//-----------------------------------------------------------------------

int StructureDomain::setConvBC(int _numConvBC, BCond *_cvbc)
{
 numConvBC = _numConvBC;
 cvbc       = _cvbc;
 return 0;
}

//-----------------------------------------------------------------------

void StructureDomain::getMatchInfo(const char *name)
{
 
  char line[500];
  FILE *fp = fopen(name, "r");

  if (!fp)  {
    fprintf(stderr, "*** ERROR: Match File \'%s\' file does not exist\n", name);
  exit(-1);
  }

  // read number of match points
  fgets(line, 500, fp);
  fgets(line, 500, fp);
  fgets(line, 500, fp);
  fgets(line, 500, fp);
  sscanf(line, "%*s %d", &numMatchData);

  // allocate memory for struct match data
  matchEle = new int[numMatchData];
  matchCoord = new double[numMatchData][2];
  gapVec = new double[numMatchData][3];
    
  // read matched elements
  int i;
  for (i = 0; i < numMatchData; i++)  {
    fgets(line, 500, fp);
    sscanf(line, "%d %lf %lf  %lf %lf %lf", 
	   matchEle+i, matchCoord[i], matchCoord[i]+1,
	   gapVec[i], gapVec[i]+1, gapVec[i]+2);
    matchEle[i] -= 1;
  }

  // close match file
  fclose(fp);
}

//-----------------------------------------------------------------------

void StructureDomain::getBinaryMatchInfo(const char *name)
{
 
}

//-----------------------------------------------------------------------

void StructureDomain::elemAdd(int elNum, int eType, int nnodes, int *n)  {

  Element *ele;

  switch(eType)
  {
    case 1:
      ele = new (ba) TwoNodeTruss(n);
      break;
    case 2:
      ele = new (ba) FourNodeQuad(n);
      // KHP: new Quad element that will work
      //      for 4, 8, 12 nodes per element
      //ele = new (ba) Quad(4, n);
      break;
    case 3:
      ele = new (ba) Therm3DQuad(n);
      break;
    case 4:
      ele = new (ba) Triangle3(n);
      break;
    case 66: // rigid beam
    case 6: {
      int nn[3] = { n[0], n[1], -1 };
      if(nnodes > 2) nn[2] = n[2];
      ele = new (ba) EulerBeam(nn);
      break;
      }
    case 7: {
      int nn[3] = { n[0], n[1], -1 };
      if(nnodes > 2) nn[2] = n[2];
      ele = new (ba) TimoshenkoBeam(nn);
      break;
      }
    case 73: // rigid 3 node shell
    case 8:
      ele = new (ba) ThreeNodeShell(n);
      break;
    case 15:
      if(nnodes == 3) ele = new (ba) ThreeNodeShell(n);
      else ele = new (ba) FourNodeShell(n);
      break;
    case 9:
      ele = new (ba) Therm2NodeBar(n);
      break;
    case 10:
      ele = new (ba) ThermQuadGal(n);
      break;
    case 11:
      ele = new (ba) TorSpring(n);
      break;
    case 17:
      ele = new (ba) EightNodeBrick(n);
      break;
    case 45:
      ele = new (ba) EightNodeBrick(n);
      break;
    case 18:
      ele = new (ba) ShearPanel(n);
      break;
    case 19:
      ele = new (ba) Membrane(n);
      break;
    case 20:
      ele = new (ba) Compo3NodeShell(n);
      break;
    case 21:
      ele = new (ba) TransSprlink(n);
      break;
    case 22:
      ele = new (ba) RotnSprlink(n);
      break;
    case 23:
      ele = new (ba) Tetrahedral(n);
      break;
    case 24:
      ele = new (ba) Pentahedral(n);
      break;
    case 25:
      ele = new (ba) TenNodeTetrahedral(n);
      break;
    case 30:
      ele = new (ba) HelmQuadGal(n);
      break;
    case 31:
      ele = new (ba) HelmQuadGls(n);
      break;
    case 35:
      ele = new (ba) HelmTri3Gal(n);
      break;
    case 36:
      ele = new (ba) HelmTri3Gls(n);
      break;
    case 40:
      ele = new (ba) TetraHelmGal(n);
      break;
    case 41:
      ele = new (ba) TetraHelmGLS(n);
      break;
    case 50:
      ele = new (ba) TetraTherm(n);
      break;
    case 52:
      ele = new (ba) ThermBrick(n);
      break;
/*
    case 60:
      ele = new (ba) HelmAxiQuad(n);
      break;
    case 61:
      ele = new (ba) HelmAxiTri(n);
      break;
*/
    case 16: // bt-shell
    case 76: // rigid four node shell
    case 88:
    case 1515:
    case 2020:
      ele = new (ba) FourNodeShell(n);
      break;
    case 95:
      ele = new (ba) HelmIsoParamHexa(nnodes,n);
      break;
    case 96:
      ele = new (ba) HelmIsoParamTetra(nnodes,n);
      break;
    default:
      cerr << "Element Type " << eType << " is not Supported." << endl;
      exit(-1);
      return;
  }

  elements[elNum] = ele;
  if(elNum >= numElem) numElem = elNum+1;

}

//-----------------------------------------------------------------------

void StructureDomain::makeConnectivity(Connectivity *elemToClus)  
{

  Domain::makeConnectivity(elemToClus);

  if (numMatchData)
    decomp->makeSubToMatch(numElem, decomp->getSubToElem(), numMatchData, matchEle);

}

//---------------------------------------------------------------------
//TODO to be deleted once numNode is not needed anymore

void StructureDomain::printGlobalConnectivity(const char *prefix)
{

  Connectivity *clusToSub = decomp->getClusToSub();
  Connectivity *subToSub = decomp->getSubToSub();
  
  char name[MAXLINE];
  sprintf(name, "%s.con", prefix);
  BinFileHandler file(name, "wb", FILE_VER);

  fprintf(stdout, "... Writing \'%s\'\n", name);

  clusToSub->write(file);
  subToSub->write(file);

  file.write(&numNode, 1);

}

//---------------------------------------------------------------------

void StructureDomain::printDecomp(const char *outPrefix)
{
  char outFileDec[MAXLINE];

  Connectivity *clusToSub = decomp->getClusToSub();
 
  int *numNeighbors = decomp->getNumNeighbors();
  int **subNeighbors = decomp->getSubNeighbors();
  Connectivity **interfNodes = decomp->getInterfNodes();

  for (int iCluster=0; iCluster < numClusters; iCluster++) {

    char *suffix = computeClusterSuffix(iCluster+1, numClusters );
    sprintf(outFileDec, "%s.dec%s", outPrefix, suffix);
    BinFileHandler outputDec(outFileDec, "wb", FILE_VER);

    fprintf(stdout, "... Writing \'%s\'\n", outFileDec);

    int numClusSubs = clusToSub->num(iCluster);

    outputDec.write(&numClusSubs, 1);
 
    BinFileHandler::OffType offsetLoc = outputDec.tell();

    BinFileHandler::OffType *subOffSets = new BinFileHandler::OffType[numClusSubs];

    outputDec.write(subOffSets, numClusSubs);

    for (int iSub = 0; iSub < numClusSubs; iSub++) {

      subOffSets[iSub] = outputDec.tell();

      int glSub = (*clusToSub)[iCluster][iSub];

      printRangesInfo(outputDec, glSub, decomp->getSubToNode(), gl2ClNodeMap[iCluster]);

      printRangesInfo(outputDec, glSub, decomp->getSubToElem(), gl2ClElemMap[iCluster]);

      printInterfaceInfo(outputDec, numNeighbors[glSub], subNeighbors[glSub], 
			 interfNodes[glSub], gl2ClNodeMap[iCluster]);

      if (numMatchData) {
	printRangesInfo(outputDec, glSub, decomp->getSubToMatch(), gl2ClMatchMap[iCluster]);
      }

    }

    outputDec.seek(offsetLoc);
    outputDec.write(subOffSets, numClusSubs);

  }

}

//---------------------------------------------------------------------

 
void StructureDomain::printStructProp(BinFileHandler &outfile)  {

  // write number of material props
  outfile.write(&numprops, 1);
  for (int i = 0; i < numprops; i++)  {

    double prop[15];

    prop[0] = sProps[i].E;
    prop[1] = sProps[i].A;
    prop[2] = sProps[i].nu;
    prop[3] = sProps[i].rho;
    prop[4] = sProps[i].eh;
    prop[5] = sProps[i].Ixx;
    prop[6] = sProps[i].Iyy;
    prop[7] = sProps[i].Izz;
    prop[8] = sProps[i].c;
    prop[9] = sProps[i].k;
    prop[10] = sProps[i].Q;
    prop[11] = sProps[i].W;
    prop[12] = sProps[i].P;
    prop[13] = sProps[i].Ta;
    prop[14] = sProps[i].kappaHelm;

    outfile.write(prop, 15);
  }
}

//------------------------------------------------------------

void StructureDomain::printCFrames(double **cf, BinFileHandler &outfile)  {

  outfile.write(&numCFrames, 1);
  for (int i = 0; i < numCFrames; i++)  {
  
    double f[9];

    for (int j = 0; j < 9; j++)
      f[j] = cf[i][j];

    outfile.write(f, 9);
  }
}
//------------------------------------------------------------

void StructureDomain::printEFrames(BinFileHandler &outfile, MapVec &gl2ClMap)  {

  int numEFrames = 0;
  BinFileHandler::OffType start = outfile.tell();
  outfile.write(&numEFrames, 1);

  EFrameData *fd = firstEFData;
  while (fd != 0)  {

    // write out global element number
    int elementNumber = gl2ClMap[fd->elnum];
    outfile.write(&elementNumber, 1);

    // write out eframe coordinates
    outfile.write(reinterpret_cast<double *>(fd->frame), 9);

   numEFrames++;
   fd = fd->next; 
  }
  BinFileHandler::OffType end = outfile.tell();
  outfile.seek(start);
  outfile.write(&numEFrames, 1);
  outfile.seek(end);
}

//------------------------------------------------------------

void StructureDomain::printAttr(int iClus, BinFileHandler &outfile,
		int nAttributes, MapVec &gl2ClElemMap)  
{

  // get connectivities for looping
  Connectivity *clusToSub = decomp->getClusToSub();
  Connectivity *subToElem = decomp->getSubToElem();
  Connectivity *elemToSub = subToElem->reverse(); // PJSA

  int locNumSubs = clusToSub->num(iClus);
  int numEle = 0;
  int iSub;

  // write out attributes per subdomain in cluster
  outfile.write(&nAttributes, 1);
  for (iSub = 0; iSub < locNumSubs; iSub++)  {

    int subD = (*clusToSub)[iClus][iSub];
/*
    numEle = subToElem->num(subD);

    // print out elements' data for current subdomain
    for (int iElem = 0; iElem < numEle; iElem++)  {
      int elemI = (*subToElem)[subD][iElem];
      
      int a[4];
      a[0] = gl2ClElemMap[attrib[elemI].nele]; // PJSA I think this is wrong, the index of attrib is not the element id
                                               // see StructureDomain::setAttrib
      a[1] = attrib[elemI].attr;
      a[2] = attrib[elemI].cmp_attr;
      a[3] = attrib[elemI].cmp_frm;
      outfile.write(a, 4);
    }
*/
    // PJSA 3-7-2007
    for(int i = 0; i < na; ++i) { // loop over all the attribs
      int elemI = attrib[i].nele;
      if((*elemToSub)[elemI][0] != subD) continue; // check if element is in current subdomain
      int a[4];
      a[0] = gl2ClElemMap[elemI];
      a[1] = attrib[i].attr;
      a[2] = attrib[i].cmp_attr;
      a[3] = attrib[i].cmp_frm;
      outfile.write(a, 4);
    }
  }
}


//---------------------------------------------------------------------

void StructureDomain::printMatchInfo(const char *outPrefix)  
{

  char outFileMatch[MAXLINE];

  Connectivity *clusToMatchElem = decomp->getClusToMatch();

  gl2ClMatchMap = new MapVec[numClusters];

  for (int iCluster = 0; iCluster < numClusters; iCluster++) {

    char *suffix = computeClusterSuffix(iCluster+1, numClusters);
    sprintf(outFileMatch, "%s.match%s", outPrefix, suffix);
    BinFileHandler outputMatch(outFileMatch, "wb", FILE_VER);

    fprintf(stdout, "... Writing \'%s\'\n", outFileMatch);

    BinFileHandler::OffType firstline = outputMatch.tell();
    outputMatch.write(&firstline, 1);
    BinFileHandler::OffType startOfData = outputMatch.tell();
    
    int numClusMatchElem = clusToMatchElem->num(iCluster);

    for (int i=0; i<numClusMatchElem; ++i) {
      int glmatch = (*clusToMatchElem)[iCluster][i];
      int glelem = matchEle[glmatch];
      gl2ClMatchMap[iCluster][glmatch] = i;

      int matchData[2] = {gl2ClElemMap[iCluster][glelem], glmatch};

      outputMatch.write(matchData, 2);
      outputMatch.write(matchCoord[glmatch], 2);
      outputMatch.write(gapVec[glmatch], 3);
    }

    BinFileHandler::OffType tocLoc = outputMatch.tell();
    outputMatch.write(&numClusMatchElem, 1);
    outputMatch.write(&startOfData, 1);
    outputMatch.seek(firstline);
    outputMatch.write(&tocLoc, 1);

  }

}

//---------------------------------------------------------------------

void StructureDomain::renumberNodes()  
{

  Domain::renumberNodes();

  // renumber BC nodes

  Connectivity *clusToSub = decomp->getClusToSub();

  initBCData();
 
  for (int iClus = 0; iClus < numClusters; iClus++)  {
    renumberBC(gl2ClNodeMap[iClus], iClus);
    if(numDbcPerClus[iClus]) 
      fprintf(stderr,"... found %d Dirichlet boundary conditions in cluster %d\n", numDbcPerClus[iClus], iClus);
  }

}

//---------------------------------------------------------------------

//void StructureDomain::initBCData()  
void Domain::initBCData()
{
  numDbcPerClus = new int[numClusters];
  numNbcPerClus = new int[numClusters];
  numCvbcPerClus = new int[numClusters];
  numIDisPerClus = new int[numClusters];
  numIDis6PerClus = new int[numClusters];
  numIvelPerClus = new int[numClusters];
  numItempPerClus = new int[numClusters];

  dbcPerClus = new int *[numClusters];
  nbcPerClus = new int *[numClusters];
  cvbcPerClus = new int *[numClusters];
  iDisPerClus = new int *[numClusters];
  iDis6PerClus = new int *[numClusters];
  iVelPerClus = new int *[numClusters];
  iTempPerClus = new int *[numClusters];

}

//---------------------------------------------------------------------

//void StructureDomain::renumberBC(MapVec &map, int iClus)  
void Domain::renumberBC(MapVec &map, int iClus)
{
  // count number of bc's in cluster
  int iBC;
  int count = 0;
  for (iBC = 0; iBC < numDirichlet; iBC++) {
    if (map.find(dbc[iBC].nnum) != map.end())
      count++;
  }
  
  numDbcPerClus[iClus] = count;
  dbcPerClus[iClus] = new int[count];
  int index = 0;
  for (iBC = 0; iBC < numDirichlet; iBC++)
    if (map.find(dbc[iBC].nnum) != map.end()) {
      //dbc[iBC].nnum = map[dbc[iBC].nnum];
      dbcPerClus[iClus][index++] = iBC; 
    }

  if (index != count)
    fprintf(stderr,"*** ERROR: Bad cluster separation of dbc's %d vs %d\n", index, count);


  count = 0;
  index = 0;
  for (iBC = 0; iBC < numNeuman; iBC++)
    if (map.find(nbc[iBC].nnum) != map.end())
      count++;

  numNbcPerClus[iClus] = count;
  nbcPerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numNeuman; iBC++)
    if (map.find(nbc[iBC].nnum) != map.end()) {
      //nbc[iBC].nnum = map[nbc[iBC].nnum];
      nbcPerClus[iClus][index++] = iBC;
    }

  count = 0;
  index = 0;
  for (iBC = 0; iBC < numConvBC; iBC++)
    if (map.find(cvbc[iBC].nnum) != map.end())
      count++;

  numCvbcPerClus[iClus] = count;
  cvbcPerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numConvBC; iBC++)
    if (map.find(cvbc[iBC].nnum) != map.end()) {
      //cvbc[iBC].nnum = map[cvbc[iBC].nnum];
      cvbcPerClus[iClus][index++] = iBC;
    }

  count = 0;
  index = 0;
  for (iBC = 0; iBC < numIDis; iBC++)
    if (map.find(iDis[iBC].nnum) != map.end())
      count++;

  numIDisPerClus[iClus] = count;
  iDisPerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numIDis; iBC++)
    if (map.find(iDis[iBC].nnum) != map.end()) {
      //iDis[iBC].nnum = map[iDis[iBC].nnum];
      iDisPerClus[iClus][index++] = iBC;
    }

  count = 0;
  index = 0;
  for (iBC = 0; iBC < numIDis6; iBC++)
    if (map.find(iDis6[iBC].nnum) != map.end())
      count++;

  numIDis6PerClus[iClus] = count;
  iDis6PerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numIDis6; iBC++)
    if (map.find(iDis6[iBC].nnum) != map.end()) {
      //iDis6[iBC].nnum = map[iDis6[iBC].nnum];
      iDis6PerClus[iClus][index++] = iBC;
    }

  count = 0;
  index = 0;
  for (iBC = 0; iBC < numIVel; iBC++)
    if (map.find(iVel[iBC].nnum) != map.end())
      count++;

  numIvelPerClus[iClus] = count;
  iVelPerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numIVel; iBC++)
    if (map.find(iVel[iBC].nnum) != map.end()) {
      //iVel[iBC].nnum = map[iVel[iBC].nnum];
      iVelPerClus[iClus][index++] = iBC;
    }

  count = 0;
  index = 0;
  for (iBC = 0; iBC < numITemp; iBC++)
    if (map.find(iTemp[iBC].nnum) != map.end())
      count++;
  
  numItempPerClus[iClus] = count;
  iTempPerClus[iClus] = new int[count];
  for (iBC = 0; iBC < numITemp; iBC++)
    if (map.find(iTemp[iBC].nnum) != map.end()) {
      //iTemp[iBC].nnum = map[iTemp[iBC].nnum];
      iTempPerClus[iClus][index++] = iBC;
    }
}

//---------------------------------------------------------------------

void StructureDomain::printGeometry(const char *outPrefix) 
{

  char outFileGeo[MAXLINE];
  char *suffix;

  // Finalize sizes of ResizeArray types
  Point3D *nodeSet = node.yield();
  Element **elems = elements.yield();
  double **compositeFrames = cframes.yield();

  Connectivity *clusToNode = decomp->getClusToNode();

  gl2ClElemMap = new MapVec[numClusters];

  for (int iCluster=0; iCluster < numClusters; iCluster++) {

    suffix = computeClusterSuffix(iCluster+1, numClusters);
    sprintf(outFileGeo, "%s.msh%s", outPrefix, suffix);

    fprintf(stdout, "... Writing \'%s\'\n", outFileGeo);

    BinFileHandler outputGeo(outFileGeo, "wb", FILE_VER);

    // save spot for ptr to start of TOC for nodes and elems
    BinFileHandler::OffType firstline = outputGeo.tell();
    outputGeo.write(&firstline, 1);
    outputGeo.write(&firstline, 1);

    // // save spot for location of start of struct. props
    outputGeo.write(&firstline, 1);

    // print out nodes
    BinFileHandler::OffType nodeListLoc = outputGeo.tell();
    printNodes(iCluster, outputGeo, nodeSet);

    // get starting location of element list
    BinFileHandler::OffType elemListLoc = outputGeo.tell();

    // print out elements
    int numClusElems;
    int *elemLoc = decomp->printElements(iCluster, outputGeo, 
		numClusElems, elems, gl2ClElemMap[iCluster]);

    BinFileHandler::OffType endOfElemsLoc = outputGeo.tell();

    // print out Material Properties for structure mesh
    printStructProp(outputGeo);
  
    // print out Attributes
    printAttr(iCluster, outputGeo, numClusElems, gl2ClElemMap[iCluster]);

    // print CFrame Data for entire domain
    printCFrames(compositeFrames, outputGeo);

    // print Eframe Data for entire domain
    printEFrames(outputGeo, gl2ClElemMap[iCluster]);
   
    // print Composite Layer Info
    printCompLayers(outputGeo);

    // print composite stiffness coefficient data
    printCoefData(outputGeo);

    // print out boundary conditions 
    printBC(outputGeo, iCluster);

    // get starting location of table of contents
    BinFileHandler::OffType tocLoc[2];
    tocLoc[0] = outputGeo.tell();     // loc of nodes TOC

    // print table of contents for nodes
    int numClusNodes = clusToNode->num(iCluster);
    fprintf(stderr,"There are %d nodes in cluster %d\n", numClusNodes, iCluster);
    outputGeo.write(&numClusNodes, 1);
    outputGeo.write(&nodeListLoc, 1);

    // print out table of contents for elements
    tocLoc[1] = outputGeo.tell();  // loc of elems TOC
    outputGeo.write(&numClusElems, 1);
    fprintf(stderr,"There are %d elems in cluster %d\n", numClusElems, iCluster);
    outputGeo.write(&elemListLoc, 1);  // start of element list
    outputGeo.write(elemLoc, numClusElems);

    // go back to beginning of file 
    outputGeo.seek(firstline);

    // write location of TOC
    outputGeo.write(tocLoc, 2);

    // write loc of start of props
    outputGeo.write(&endOfElemsLoc, 1);

  }
}

//---------------------------------------------------------------------

int StructureDomain::addMat(int nmat, StructProp &p)
{

 if(nmat >= numprops) {
    int n_max= (3*numprops)/2;
    if(nmat>=n_max) n_max = nmat+1;
    StructProp *n_matTable = new StructProp[n_max];
    int i;
    for(i=0; i < numprops; ++i)
      n_matTable[i] = sProps[i];
    delete[]sProps;
    sProps = n_matTable;
    numprops = n_max;
 }

 sProps[nmat] = p;
 return 0;
}

//---------------------------------------------------------------------

int StructureDomain::addLay(int l, LayInfo *linfo)
{
  layInfo[l] = linfo;
  int newnum = l+1;

  if (numlayInfo < newnum)
    numlayInfo = newnum;

  return 0;
}

//---------------------------------------------------------------------

int StructureDomain::addCoefInfo(int cin, CoefData &cdata)  {

  CoefData *cd  = new CoefData(cdata);
  coefData[cin] = cd;
  if (numCoefData < cin) 
    numCoefData = cin;
  return 0;
}

//---------------------------------------------------------------------

int StructureDomain::setAttrib(int n, int a, int ca, int cfrm)
{
  if(na == namax) {
    int nmx = (3*namax)/2;
    Attrib *n_attrib = new Attrib[nmx];
    int i;
    for(i=0; i < namax; ++i)
      n_attrib[i] = attrib[i];
    delete[] attrib;
    namax = nmx;
    attrib = n_attrib;
  }

  attrib[na].nele = n;
  attrib[na].attr = a;
  attrib[na].cmp_attr = ca;
  attrib[na].cmp_frm  = cfrm;
  na++;

  return 0;
}

//---------------------------------------------------------------------

int StructureDomain::setFrame(int el, double *data)  {

  EFrameData *efd = new EFrameData;
  efd->next = firstEFData;
  firstEFData = efd;
  efd->elnum = el;
  int i,j;
  for(i =0; i < 3; ++i)
    for(j =0; j < 3; ++j)
      efd->frame[i][j] = data[i*3+j];
  return 0;
}

//---------------------------------------------------------------------

int StructureDomain::addCFrame(int fn, double *f)  {

  if(fn >= numCFrames)
    numCFrames = fn+1;

  cframes[fn] = new double[9];

  int i;
  for(i=0; i < 9; ++i)
    cframes[fn][i] = f[i];

  return 0;
}


//----------------------------------------------------------------

int StructureDomain::setIDis6(int _numIDis6, BCond *_iDis6)
{
  numIDis6 = _numIDis6;
  iDis6    = _iDis6;
  return 0;
}

//----------------------------------------------------------------

int StructureDomain::setIDis(int _numIDis, BCond *_iDis)
{
  numIDis = _numIDis;
  iDis    = _iDis;
  return 0;
}

//----------------------------------------------------------------

void StructureDomain::printCompLayers(BinFileHandler &outfile)  {

  outfile.write(&numlayInfo, 1);
  for (int i = 0; i < numlayInfo; i++)  {

    int intData[3];
    intData[0] = layInfo[i]->numLayers;
    intData[1] = layInfo[i]->maxLayer;
    intData[2] = layInfo[i]->type;

    double d[9];

    for (int iLay = 0; iLay < intData[0]; iLay++)
      for (int j = 0; j < 9; j++)
        d[j] = layInfo[i]->data[iLay][j];

    outfile.write(intData, 3);
    outfile.write(d, 9*intData[0]);
  }
}

//----------------------------------------------------------------

void StructureDomain::printCoefData(BinFileHandler &outfile)  {

  outfile.write(&numCoefData, 1);

  for (int i = 0; i < numCoefData; i++)  {

    double *data = coefData[i]->values();
    outfile.write(data, 36);
  }

}

//----------------------------------------------------------------
//void StructureDomain::printBC(BinFileHandler &outfile, int iClus)  
void Domain::printBC(BinFileHandler &outfile, int iClus)
{

  int iBC;

/*
  fprintf(stderr,"numDirichlet: %d\n", numDbcPerClus[iClus]);
  fprintf(stderr,"numNeuman: %d\n", numNbcPerClus[iClus]);
  fprintf(stderr,"numConvective: %d\n", numCvbcPerClus[iClus]);
  fprintf(stderr,"numIDisp: %d\n", numIDisPerClus[iClus]);
  fprintf(stderr,"numIDisp6: %d\n", numIDis6PerClus[iClus]);
  fprintf(stderr,"numIVel: %d\n", numIvelPerClus[iClus]);
  fprintf(stderr,"numITemp: %d\n", numItempPerClus[iClus]);
*/
 
  // print Dirichlet BC's
  outfile.write(&numDbcPerClus[iClus], 1);
  for (iBC = 0; iBC < numDbcPerClus[iClus]; iBC++)  {
    int bcNode = dbcPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][dbc[bcNode].nnum], 1);
    outfile.write(&dbc[bcNode].dofnum, 1);
    outfile.write(&dbc[bcNode].val, 1);
  }

  // print Neuman BC's
  outfile.write(&numNbcPerClus[iClus], 1);
  for (iBC = 0; iBC < numNbcPerClus[iClus]; iBC++)  {
    int bcNode = nbcPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][nbc[bcNode].nnum], 1);
    outfile.write(&nbc[bcNode].dofnum, 1);
    outfile.write(&nbc[bcNode].val, 1);
  }

  // print Convective BC's
  outfile.write(&numConvBC, 1);
  for (iBC = 0; iBC < numCvbcPerClus[iClus]; iBC++)  {
    int bcNode = cvbcPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][cvbc[bcNode].nnum], 1);
    outfile.write(&cvbc[bcNode].dofnum, 1);
    outfile.write(&cvbc[bcNode].val, 1);
  }

  // print Initial Displacement BC's
  outfile.write(&numIDisPerClus[iClus], 1);
  for (iBC = 0; iBC < numIDisPerClus[iClus]; iBC++)  {
    int bcNode = iDisPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][iDis[bcNode].nnum], 1);
    outfile.write(&iDis[bcNode].dofnum, 1);
    outfile.write(&iDis[bcNode].val, 1);
  }

  // print Initial Displacement BC's  (6 colmumn format)
  outfile.write(&numIDis6PerClus[iClus], 1);

  for (iBC = 0; iBC < numIDis6PerClus[iClus]; iBC++)  {
    int bcNode = iDis6PerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][iDis6[bcNode].nnum], 1);
    outfile.write(&iDis6[bcNode].dofnum, 1);
    outfile.write(&iDis6[bcNode].val, 1);
  }

  // print Initial Velocity BC's
  outfile.write(&numIvelPerClus[iClus], 1);
  for (iBC = 0; iBC < numIvelPerClus[iClus]; iBC++)  {
    int bcNode = iVelPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][iVel[bcNode].nnum], 1);
    outfile.write(&iVel[bcNode].dofnum, 1);
    outfile.write(&iVel[bcNode].val, 1);
  }

  // print Initial Temperature BC's 
  outfile.write(&numItempPerClus[iClus], 1);

  for (iBC = 0; iBC < numItempPerClus[iClus]; iBC++)  {
    int bcNode = iTempPerClus[iClus][iBC];
    outfile.write(&gl2ClNodeMap[iClus][iTemp[bcNode].nnum], 1);
    outfile.write(&iTemp[bcNode].dofnum, 1);
    outfile.write(&iTemp[bcNode].val, 1);
  }
}


//------------------------------------------------------------

